#!/usr/bin/env python
# Copyright 2014-2023 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Matthew Hennefarth <mhennefarth@uchicago.com>

from pyscf.grad import rks as rks_grad
from pyscf.dft import gen_grid
from pyscf.lib import logger, tag_array, pack_tril, current_memory
from pyscf.mcscf import casci, mc1step, newton_casscf
from pyscf.grad import sacasscf
from pyscf.mcscf.casci import cas_natorb

from pyscf.mcpdft.otpd import get_ontop_pair_density, _grid_ao2mo
from pyscf.mcpdft.tfnal_derivs import contract_fot, unpack_vot, contract_vot
from pyscf.mcpdft import _dms
from pyscf.mcpdft import mspdft
import pyscf.grad.mcpdft as mcpdft_grad

import numpy as np
import gc

BLKSIZE = gen_grid.BLKSIZE


def get_ontop_response(
    mc, ot, state, atmlst, casdm1, casdm1_0, mo_coeff=None, ci=None, max_memory=None
):
    if mo_coeff is None:
        mo_coeff = mc.mo_coeff
    if ci is None:
        ci = mc.ci
    if max_memory is None:
        max_memory = mc.max_memory

    t0 = (logger.process_clock(), logger.perf_counter())

    mol = mc.mol
    ncore = mc.ncore
    ncas = mc.ncas
    nocc = ncore + ncas
    nao, nmo = mo_coeff.shape

    dvxc = np.zeros((3, nao))
    de_grid = np.zeros((len(atmlst), 3))
    de_wgt = np.zeros((len(atmlst), 3))

    mo_coeff_0, ci_0, mo_occup_0 = cas_natorb(
        mc, mo_coeff=mo_coeff, ci=ci, casdm1=casdm1_0
    )
    mo_coeff, ci, mo_occup = cas_natorb(mc, mo_coeff=mo_coeff, ci=ci, casdm1=casdm1)

    mo_occ = mo_coeff[:, :nocc]
    mo_cas = mo_coeff[:, ncore:nocc]

    mo_occ_0 = mo_coeff_0[:, :nocc]
    mo_cas_0 = mo_coeff_0[:, ncore:nocc]

    # Need to regenerate these with the updated ci values....
    casdm1s = mc.make_one_casdm1s(ci=ci, state=state)
    casdm1 = casdm1s[0] + casdm1s[1]
    casdm2 = mc.make_one_casdm2(ci=ci, state=state)
    dm1s = _dms.casdm1s_to_dm1s(mc, casdm1s, mo_coeff=mo_coeff, ncore=ncore, ncas=ncas)
    dm1 = dm1s[0] + dm1s[1]
    dm1 = tag_array(dm1, mo_coeff=mo_coeff[:, :nocc], mo_occ=mo_occup[:nocc])

    casdm1s_0, casdm2_0 = mc.get_casdm12_0(ci=ci_0)
    casdm1_0 = casdm1s_0[0] + casdm1s_0[1]
    dm1s_0 = _dms.casdm1s_to_dm1s(
        mc, casdm1s_0, mo_coeff=mo_coeff_0, ncore=ncore, ncas=ncas
    )
    dm1_0 = dm1s_0[0] + dm1s_0[1]
    dm1_0 = tag_array(dm1_0, mo_coeff=mo_coeff_0[:, :nocc], mo_occ=mo_occup_0[:nocc])

    cascm2 = _dms.dm2_cumulant(casdm2, casdm1)
    cascm2_0 = _dms.dm2_cumulant(casdm2_0, casdm1_0)

    make_rho = ot._numint._gen_rho_evaluator(mol, dm1, 1)[0]
    make_rho_0 = ot._numint._gen_rho_evaluator(mol, dm1_0, 1)[0]

    idx = np.array([[1, 4, 5, 6], [2, 5, 7, 8], [3, 6, 8, 9]], dtype=np.int_)
    # For addressing particular ao derivatives
    if ot.xctype == "LDA":
        idx = idx[:, 0:1]  # For LDAs, no second derivatives

    casdm2_pack = mcpdft_grad.pack_casdm2(cascm2, ncas)
    casdm2_0_pack = mcpdft_grad.pack_casdm2(cascm2_0, ncas)

    full_atmlst = -np.ones(mol.natm, dtype=np.int_)
    for k, ia in enumerate(atmlst):
        full_atmlst[ia] = k

    t1 = logger.timer(mc, "L-PDFT HlFn quadrature setup", *t0)

    ndao = (1, 4)[ot.dens_deriv]
    ndpi = (1, 4)[ot.Pi_deriv]
    ncols = 1.05 * 3 * (ndao * nao + nocc) + max(ndao * nao, ndpi * ncas * ncas)
    # I have no idea if this is actually the correct number of columns, but I have a feeling it is not since I should
    # be accounting for the extra rows from feff stuff...

    for ia, (coords, w0, w1) in enumerate(rks_grad.grids_response_cc(ot.grids)):
        gc.collect()
        ngrids = coords.shape[0]
        remaining_floats = (max_memory - current_memory()[0]) * 1e6 / 8
        blksize = int(remaining_floats / (ncols * BLKSIZE)) * BLKSIZE
        blksize = max(BLKSIZE, min(blksize, ngrids, BLKSIZE * 1200))
        t1 = logger.timer(
            mc, "L-PDFT HlFn quadrature atom {} mask and memory setup".format(ia), *t1
        )
        for ip0 in range(0, ngrids, blksize):
            ip1 = min(ngrids, ip0 + blksize)
            mask = gen_grid.make_mask(mol, coords[ip0:ip1])
            logger.info(
                mc,
                "L-PDFT gradient atom {} slice {}-{} of {} total".format(
                    ia, ip0, ip1, ngrids
                ),
            )
            ao = ot._numint.eval_ao(
                mol, coords[ip0:ip1], deriv=ot.dens_deriv + 1, non0tab=mask
            )

            t1 = logger.timer(
                mc, "L-PDFT HlFn quadrature atom {} ao grids".format(ia), *t1
            )

            if ot.xctype == "LDA":
                aoval = ao[0]

            if ot.xctype == "GGA":
                aoval = ao[:4]

            rho = make_rho(0, aoval, mask, ot.xctype) / 2.0
            rho = np.stack((rho,) * 2, axis=0)
            rho_0 = make_rho_0(0, aoval, mask, ot.xctype) / 2.0
            rho_0 = np.stack((rho_0,) * 2, axis=0)
            delta_rho = rho - rho_0
            t1 = logger.timer(
                mc, "L-PDFT HlFn quadrature atom {} rho calc".format(ia), *t1
            )

            Pi = get_ontop_pair_density(
                ot, rho, aoval, cascm2, mo_cas, ot.dens_deriv, mask
            )
            Pi_0 = get_ontop_pair_density(
                ot, rho_0, aoval, cascm2_0, mo_cas_0, ot.dens_deriv, mask
            )
            delta_Pi = Pi - Pi_0
            t1 = logger.timer(
                mc, "L-PDFT HlFn quadrature atom {} Pi calc".format(ia), *t1
            )

            if ot.xctype == "LDA":
                aoval = ao[:1]

            moval_occ = _grid_ao2mo(mol, aoval, mo_occ, mask)
            moval_occ_0 = _grid_ao2mo(mol, aoval, mo_occ_0, mask)
            t1 = logger.timer(
                mc, "L-PDFT HlFn quadrature atom {} ao2mo grids".format(ia), *t1
            )

            aoval = np.ascontiguousarray(
                [ao[ix].transpose(0, 2, 1) for ix in idx[:, :ndao]]
            ).transpose(0, 1, 3, 2)
            ao = None
            t1 = logger.timer(
                mc, "L-PDFT HlFn quadrature atom {} ao grid reshape".format(ia), *t1
            )

            eot, vot, fot = ot.eval_ot(
                rho_0, Pi_0, weights=w0[ip0:ip1], dderiv=2, _unpack_vot=False
            )
            fot = contract_fot(
                ot, fot, rho_0, Pi_0, delta_rho, delta_Pi, unpack=True, vot_packed=vot
            )
            vot = unpack_vot(vot, rho_0, Pi_0)
            # See the equations...
            eot += contract_vot(vot, delta_rho, delta_Pi)
            t1 = logger.timer(
                mc, "PDFT HlFn quadrature atom {} eval_ot".format(ia), *t1
            )

            puvx_mem = 2 * ndpi * (ip1 - ip0) * ncas * ncas * 8 / 1e6
            remaining_mem = max_memory - current_memory()[0]
            logger.info(
                mc,
                (
                    "L-PDFT gradient memory note: working on {} grid points: estimated puvx usage = {:.1f} of {:.1f} "
                    "remaining MB"
                ).format((ip1 - ip0), puvx_mem, remaining_mem),
            )

            # Weight response
            de_wgt += np.tensordot(eot, w1[atmlst, ..., ip0:ip1], axes=(0, 2))
            t1 = logger.timer(
                mc, "L-PDFT HlFn quadrature atom {} weight response".format(ia), *t1
            )

            # The mo_occup values might be screwing me here...
            # rho_delta * fot * drho_SA/dx
            tmp_df = mcpdft_grad.xc_response(
                ot,
                fot,
                rho_0,
                Pi_0,
                w0[ip0:ip1],
                moval_occ_0,
                aoval,
                mo_occ_0,
                mo_occup_0,
                ncore,
                nocc,
                casdm2_0_pack,
                ndpi,
                mo_cas_0,
            )
            # vot * drho_Gamma/dx
            tmp_dv = mcpdft_grad.xc_response(
                ot,
                vot,
                rho,
                Pi,
                w0[ip0:ip1],
                moval_occ,
                aoval,
                mo_occ,
                mo_occup,
                ncore,
                nocc,
                casdm2_pack,
                ndpi,
                mo_cas,
            )

            tmp_dxc = tmp_df + tmp_dv

            # Find the atoms that are part of the atomlist
            # grid correction shouldn't be added if they arent there
            k = full_atmlst[ia]
            if k >= 0:
                de_grid[k] += 2 * tmp_dxc.sum(1)  # Grid response

            dvxc -= tmp_dxc  # XC response

            tmp_dxc = tmp_df = tmp_dv = None
            t1 = logger.timer(mc, "L-PDFT HlFn quadrature atom {}".format(ia), *t1)

            rho_0 = Pi_0 = rho = Pi = delta_rho = delta_Pi = None
            eot = vot = fot = aoval = moval_occ = moval_occ_0 = None
            gc.collect()

    return dvxc, de_wgt, de_grid


def lpdft_HellmanFeynman_grad(
    mc,
    ot,
    state,
    feff1,
    feff2,
    mo_coeff=None,
    ci=None,
    atmlst=None,
    mf_grad=None,
    verbose=None,
    max_memory=None,
    auxbasis_response=False,
):
    if mo_coeff is None:
        mo_coeff = mc.mo_coeff
    if ci is None:
        ci = mc.ci
    if mf_grad is None:
        mf_grad = mc.get_rhf_base ().nuc_grad_method()
    if mc.frozen is not None:
        raise NotImplementedError
    mol = mc.mol
    if atmlst is None:
        atmlst = range(mol.natm)

    t0 = (logger.process_clock(), logger.perf_counter())

    # Specific state density
    casdm1s = mc.make_one_casdm1s(ci=ci, state=state)
    casdm1 = casdm1s[0] + casdm1s[1]
    dm1s = _dms.casdm1s_to_dm1s(mc, casdm1s=casdm1s, mo_coeff=mo_coeff)
    dm1 = dm1s[0] + dm1s[1]
    casdm2 = mc.make_one_casdm2(ci=ci, state=state)

    # The model-space density (or state-average density)
    casdm1s_0, casdm2_0 = mc.get_casdm12_0()
    dm1s_0 = _dms.casdm1s_to_dm1s(mc, casdm1s=casdm1s_0, mo_coeff=mo_coeff)
    dm1_0 = dm1s_0[0] + dm1s_0[1]
    casdm1_0 = casdm1s_0[0] + casdm1s_0[1]

    # Generate the Generalized Fock Component
    gfock_expl = mcpdft_grad.gfock_sym(
        mc, mo_coeff, casdm1, casdm2, mc.get_lpdft_hcore(), mc.veff2
    )
    gfock_impl = mcpdft_grad.gfock_sym(mc, mo_coeff, casdm1_0, casdm2_0, feff1, feff2)
    gfock = gfock_expl + gfock_impl

    dme0 = mo_coeff @ (0.5 * (gfock + gfock.T)) @ mo_coeff.T
    del gfock, gfock_impl, gfock_expl
    t0 = logger.timer(mc, "L-PDFT HlFn gfock", *t0)

    # Coulomb potential derivatives generated from zero-order density
    dvj_all = mf_grad.get_j(dm=[dm1, dm1_0])
    dvj = dvj_all[0]
    dvj_0 = dvj_all[1]

    dvxc, de_wgt, de_grid = get_ontop_response(
        mc,
        ot,
        state,
        atmlst,
        casdm1,
        casdm1_0,
        mo_coeff=mo_coeff.copy(),
        ci=ci.copy(),
        max_memory=max_memory,
    )

    delta_dm1 = dm1 - dm1_0

    def coul_term(p0, p1):
        return 2 * (
            np.tensordot(dvj_0[:, p0:p1], delta_dm1[p0:p1])
            + np.tensordot(dvj[:, p0:p1], dm1_0[p0:p1])
        )

    de_hcore, de_coul, de_xc, de_nuc, de_renorm = mcpdft_grad.sum_terms(
        mf_grad, mol, atmlst, dm1, dme0, coul_term, dvxc
    )

    logger.debug(mc, "L-PDFT Hellmann-Feynman nuclear:\n{}".format(de_nuc))
    logger.debug(mc, "L-PDFT Hellmann-Feynman hcore component:\n{}".format(de_hcore))
    logger.debug(mc, "L-PDFT Hellmann-Feynman coulomb component:\n{}".format(de_coul))
    logger.debug(mc, "L-PDFT Hellmann-Feynman xc component:\n{}".format(de_xc))
    logger.debug(
        mc, "L-PDFT Hellmann-Feynman quadrature point component:\n{}".format(de_grid)
    )
    logger.debug(
        mc, "L-PDFT Hellmann-Feynman quadrature weight component:\n{}".format(de_wgt)
    )
    logger.debug(mc, "L-PDFT Hellmann-Feynman renorm component:\n{}".format(de_renorm))

    de = de_nuc + de_hcore + de_coul + de_renorm + de_xc + de_grid + de_wgt

    if auxbasis_response:
        dvj_aux = dvj_all.aux[:,:,atmlst,:]
        de_aux = dvj_aux[1, 0] + dvj_aux[0, 1] - dvj_aux[1, 1]
        logger.debug(mc, "L-PDFT Hellmann-Feynman aux component:\n{}".format(de_aux))
        de += de_aux

    logger.timer(mc, "L-PDFT HlFn total", *t0)

    return de


class Gradients(sacasscf.Gradients):
    def __init(self, pdft, state=None):
        super().__init__(pdft, state=state)

        if self.state is None and self.nroots == 1:
            self.state = 0

        self.e_mcscf = self.base.e_mcscf
        self._not_implemented_check()

    def _not_implemented_check(self):
        name = self.__class__.__name__
        if isinstance(self.base, casci.CASCI) and not isinstance(
            self.base, mc1step.CASSCF
        ):
            raise NotImplementedError("{} for CASCI-based MC-PDFT".format(name))
        ot, otxc, nelecas = self.base.otfnal, self.base.otxc, self.base.nelecas
        spin = abs(nelecas[0] - nelecas[1])
        omega, alpha, hyb = ot._numint.rsh_and_hybrid_coeff(otxc, spin=spin)
        hyb_x, hyb_c = hyb
        if hyb_x or hyb_c:
            raise NotImplementedError("{} for hybrid MC-PDFT functionals".format(name))
        if omega:
            raise NotImplementedError(
                "{} for range-separated MC-PDFT functionals".format(name)
            )

    def kernel(self, **kwargs):
        state = kwargs["state"] if "state" in kwargs else self.state
        if state is None:
            raise NotImplementedError("Gradient of LPDFT state-average energy")
        self.state = state
        mo = kwargs["mo"] if "mo" in kwargs else self.base.mo_coeff
        ci = kwargs["ci"] if "ci" in kwargs else self.base.ci
        if isinstance(ci, np.ndarray):
            ci = [ci]  # hack hack hack????? idk
        kwargs["ci"] = ci
        # need to compute feff1, feff2 if not already in kwargs
        if ("feff1" not in kwargs) or ("feff2" not in kwargs):
            kwargs["feff1"], kwargs["feff2"] = self.get_otp_gradient_response(
                mo, ci, state
            )

        return super().kernel(**kwargs)

    def get_wfn_response(
        self,
        state=None,
        verbose=None,
        mo=None,
        ci=None,
        feff1=None,
        feff2=None,
        **kwargs
    ):
        """Returns the derivative of the L-PDFT energy for the given state with respect to MO parameters and CI
        parameters. Care is take to account for the implicit and explicit response terms, and make sure the CI
        vectors are properly projected out.

        Args:
            state : int
                Which state energy to get response of.

            mo : ndarray of shape (nao, nmo)
                A full set of molecular orbital coefficients. Taken from self if not provided.

            ci : list of ndarrays of length nroots
                CI vectors should be from a converged L-PDFT calculation.

            feff1 : ndarray of shape (nao, nao) 1-particle On-top gradient response which as been contracted with the
                Delta density generated from state. Should include the Coulomb term as well.

            feff2 : pyscf.mcscf.mc_ao2mo._ERIS instance Relevant 2-body on-top gradient response terms in the MO
                basis. Also, partially contracted with the Delta density.

        Returns: g_all : ndarray of shape nlag First sector [:self.ngorb] contains the derivatives with respect to MO
        parameters. Second sector [self.ngorb:] contains the derivatives with respect to CI parameters.
        """
        if state is None:
            state = self.state

        if verbose is None:
            verbose = self.verbose

        if mo is None:
            mo = self.base.mo_coeff

        if ci is None:
            ci = self.base.ci

        if verbose is None:
            verbose = self.verbose

        if (feff1 is None) or (feff2 is None):
            feff1, feff2 = self.get_otp_gradient_response(mo, ci, state)

        log = logger.new_logger(self, verbose)

        ndet = self.na_states[state] * self.nb_states[state]
        fcasscf = self.make_fcasscf(state)

        # Exploit (hopefully) the fact that the zero-order density is
        # really just the State Average Density!
        fcasscf_sa = self.make_fcasscf_sa()

        fcasscf.mo_coeff = mo
        fcasscf.ci = ci[state]

        fcasscf.get_hcore = self.base.get_lpdft_hcore
        fcasscf_sa.get_hcore = lambda: feff1

        g_all_explicit = newton_casscf.gen_g_hop(
            fcasscf, mo, ci[state], self.base.veff2, verbose
        )[0]
        g_all_implicit = newton_casscf.gen_g_hop(fcasscf_sa, mo, ci, feff2, verbose)[0]

        # Debug
        log.debug("g_all explicit orb:\n{}".format(g_all_explicit[: self.ngorb]))
        log.debug("g_all explicit ci:\n{}".format(g_all_explicit[self.ngorb :]))
        log.debug("g_all implicit orb:\n{}".format(g_all_implicit[: self.ngorb]))
        log.debug("g_all implicit ci:\n{}".format(g_all_implicit[self.ngorb :]))

        # Need to remove the SA-SA rotations from g_all_implicit CI contributions
        spin_states = np.asarray(self.spin_states)
        gmo_implicit, gci_implicit = self.unpack_uniq_var(g_all_implicit)
        for root in range(self.nroots):
            idx_spin = spin_states == spin_states[root]
            idx = np.where(idx_spin)[0]

            gci_root = gci_implicit[root].ravel()

            assert root in idx
            ci_proj = np.asarray([ci[i].ravel() for i in idx])
            gci_sa = np.dot(ci_proj, gci_root)
            gci_root -= np.dot(gci_sa, ci_proj)

            gci_implicit[root] = gci_root

        g_all = self.pack_uniq_var(gmo_implicit, gci_implicit)

        g_all[: self.ngorb] += g_all_explicit[: self.ngorb]
        offs = (
            sum(
                [
                    na * nb
                    for na, nb in zip(self.na_states[:state], self.nb_states[:state])
                ]
            )
            if root > 0
            else 0
        )
        g_all[self.ngorb :][offs:][:ndet] += g_all_explicit[self.ngorb :]

        gorb, gci = self.unpack_uniq_var(g_all)
        log.debug("g_all orb:\n{}".format(gorb))
        log.debug("g_all ci:\n{}".format([c.ravel() for c in gci]))

        return g_all

    def get_ham_response(
        self,
        state=None,
        atmlst=None,
        verbose=None,
        mo=None,
        ci=None,
        mf_grad=None,
        feff1=None,
        feff2=None,
        **kwargs
    ):
        if state is None:
            state = self.state

        if atmlst is None:
            atmlst = self.atmlst

        if verbose is None:
            verbose = self.verbose

        if mo is None:
            mo = self.base.mo_coeff

        if ci is None:
            ci = self.base.ci

        if (feff1 is None) or (feff2 is None):
            assert False, kwargs

        return lpdft_HellmanFeynman_grad(
            self.base,
            self.base.otfnal,
            state,
            feff1=feff1,
            feff2=feff2,
            mo_coeff=mo,
            ci=ci,
            atmlst=atmlst,
            mf_grad=mf_grad,
            verbose=verbose,
        )

    def get_otp_gradient_response(self, mo=None, ci=None, state=0):
        """Generate the 1- and 2-body on-top gradient response terms which have been partially contracted with the
        Delta density generated from state.

        Args:
            mo : ndarray of shape (nao,nmo)
                A full set of molecular orbital coefficients. Taken from self if not provided.

            ci : list of ndarrays of length nroots
                CI vectors should be from a converged L-PDFT calculation

            state : int
                State to generate the Delta density with

        Returns:
            feff1 : ndarray of shape (nao, nao) 1-particle On-top gradient response which as been contracted with the
                Delta density generated from state. Should include the Coulomb term as well.

            feff2 : pyscf.mcscf.mc_ao2mo._ERIS instance Relevant 2-body on-top gradient response terms in the MO
                basis. Also, partially contracted with the Delta density.
        """
        if mo is None:
            mo = self.base.mo_coeff

        if ci is None:
            ci = self.base.ci

        if state is None:
            state = self.state

        # This is the zero-order density
        casdm1s_0, casdm2_0 = self.base.get_casdm12_0()

        # This is the density of the state we are differentiating with respect to
        casdm1s = self.base.make_one_casdm1s(ci=ci, state=state)
        casdm2 = self.base.make_one_casdm2(ci=ci, state=state)
        dm1s = _dms.casdm1s_to_dm1s(self.base, casdm1s, mo_coeff=mo)

        cascm2 = _dms.dm2_cumulant(casdm2, casdm1s)

        return self.base.get_pdft_feff(
            mo=mo,
            ci=ci,
            casdm1s=casdm1s_0,
            casdm2=casdm2_0,
            c_dm1s=dm1s,
            c_cascm2=cascm2,
            jk_pc=True,
            paaa_only=True,
            incl_coul=True,
            delta=True,
        )

    def get_Aop_Adiag(
        self, verbose=None, mo=None, ci=None, eris=None, state=None, **kwargs
    ):
        """This function accounts for the fact that the CI vectors are no longer eigenstates of the CAS Hamiltonian.
        It adds back in the necessary values to the Hessian."""
        if verbose is None:
            verbose = self.verbose

        if mo is None:
            mo = self.base.mo_coeff

        if ci is None:
            ci = self.base.ci

        if state is None:
            state = self.state

        if eris is None and self.eris is None:
            eris = self.eris = self.base.ao2mo(mo)

        elif eris is None:
            eris = self.eris

        ham_od = mspdft.make_heff_mcscf(self.base, mo_coeff=mo, ci=ci)
        fcasscf = self.make_fcasscf_sa()

        hop, Adiag = newton_casscf.gen_g_hop(fcasscf, mo, ci, eris, verbose)[2:]

        # TODO: (MR Hennefarth) This is not safe way to check if something is a
        # mixed solver, can probably do better than this...
        if hasattr(self.base, "_irrep_slices"):
            for ham_slice in ham_od:
                ham_slice[np.diag_indices_from(ham_slice)] = 0.0
                ham_slice += (
                    ham_slice.T
                )  # This corresponds to the arbitrary newton_casscf*2

            def Aop(x):
                Ax = hop(x)
                x_c = self.unpack_uniq_var(x)[1]
                Ax_o, Ax_c = self.unpack_uniq_var(Ax)

                for irrep_slice, ham_slice in zip(self.base._irrep_slices, ham_od):
                    Ax_c_od_slice = list(
                        np.tensordot(
                            -ham_slice, np.stack(x_c[irrep_slice], axis=0), axes=1
                        )
                    )
                    Ax_c[irrep_slice] = [
                        a1 + (w * a2)
                        for a1, a2, w in zip(
                            Ax_c[irrep_slice],
                            Ax_c_od_slice,
                            self.base.weights[irrep_slice],
                        )
                    ]

                return self.pack_uniq_var(Ax_o, Ax_c)

        else:
            ham_od[np.diag_indices_from(ham_od)] = 0.0
            ham_od += ham_od.T  # This corresponds to the arbitrary newton_casscf*2

            def Aop(x):
                Ax = hop(x)
                x_c = self.unpack_uniq_var(x)[1]
                Ax_o, Ax_c = self.unpack_uniq_var(Ax)
                Ax_c_od = list(np.tensordot(-ham_od, np.stack(x_c, axis=0), axes=1))
                Ax_c = [
                    a1 + (w * a2) for a1, a2, w in zip(Ax_c, Ax_c_od, self.base.weights)
                ]
                return self.pack_uniq_var(Ax_o, Ax_c)

        return self.project_Aop(Aop, ci, state), Adiag



#!/usr/bin/env python
# Copyright 2022-2023 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Authors: Xing Zhang <zhangxing.nju@gmail.com>
#

import numpy as np
from pyscf import lib
from pyscf.pbc.lib import ktensor

einsum = lib.einsum

def cc_Foo(kpts, kqrts, t1, t2, eris, rmat):
    nkpts, nocc, nvir = t1.shape
    #Fki = np.empty((nkpts,nocc,nocc), dtype=t2.dtype)
    metadata = {'kpts': kpts, 'rmat': rmat,
                'label': 'oo', 'trans': 'cn',
                'incore': True}
    Fki = ktensor.empty([nocc,nocc], dtype=t2.dtype, metadata=metadata)

    for i in range(kpts.nkpts_ibz):
        ki = kpts.ibz2bz[i]
        Fki[ki] = eris.fock[ki,:nocc,:nocc].copy()

    for i, kq in enumerate(kqrts.kqrts_ibz):
        ki, kl, kc, kd = kq
        kk = ki
        Soovv = 2 * eris.oovv[kk,kl,kc] - eris.oovv[kk,kl,kd].transpose(0,1,3,2)
        fock = einsum('klcd,ilcd->ki', Soovv, t2[ki,kl,kc])
        if ki == kc:
            fock += einsum('klcd,ic,ld->ki', Soovv, t1[ki], t1[kl])
        for _, iop in kqrts.loop_stabilizer(i):
            rmat_oo = rmat.oo[ki][iop]
            Fki[ki] += einsum('ki,km,in->mn', fock, rmat_oo.conj(), rmat_oo)
    return Fki

def cc_Fov(kpts, kqrts, t1, t2, eris, rmat):
    nkpts, nocc, nvir = t1.shape
    #Fkc = np.empty((nkpts,nocc,nvir), dtype=t2.dtype)
    metadata = {'kpts': kpts, 'rmat': rmat,
                'label': 'ov', 'trans': 'cn',
                'incore': True}
    Fkc = ktensor.empty([nocc,nvir], dtype=t2.dtype, metadata=metadata)

    for i in range(kpts.nkpts_ibz):
        ki = kpts.ibz2bz[i]
        Fkc[ki] = eris.fock[ki,:nocc,nocc:].copy()

    for i, kq in enumerate(kqrts.kqrts_ibz):
        kk, kl, kc, kd = kq
        if kc == kk and kl == kd:
            Soovv = 2 * eris.oovv[kk,kl,kk] - eris.oovv[kk,kl,kl].transpose(0,1,3,2)
            fock = einsum('klcd,ld->kc', Soovv, t1[kl])
            for _, iop in kqrts.loop_stabilizer(i):
                rmat_oo = rmat.oo[kk][iop]
                rmat_vv = rmat.vv[kk][iop]
                Fkc[kk] += einsum('kc,km,cb->mb', fock, rmat_oo.conj(), rmat_vv)
    return Fkc

def cc_Fvv(kpts, kqrts, t1, t2, eris, rmat):
    nkpts, nocc, nvir = t1.shape
    #Fac = np.empty((nkpts,nvir,nvir), dtype=t2.dtype)
    metadata = {'kpts': kpts, 'rmat': rmat,
                'label': 'vv', 'trans': 'cn',
                'incore': True}
    Fac = ktensor.empty([nvir,nvir], dtype=t2.dtype, metadata=metadata)

    for i in range(kpts.nkpts_ibz):
        ki = kpts.ibz2bz[i]
        Fac[ki] = eris.fock[ki,nocc:,nocc:].copy()

    ka_ibz_bz = kpts.ibz2bz[np.arange(kpts.nkpts_ibz)]
    for i, kq in enumerate(kqrts.kqrts_ibz):
        kk, kl, ka, kd = kq
        kc = ka
        Soovv = 2*eris.oovv[kk,kl,kc] - eris.oovv[kk,kl,kd].transpose(0,1,3,2)
        fock = -einsum('klcd,klad->ac', Soovv, t2[kk,kl,ka])
        if kk == ka:
            fock += -einsum('klcd,ka,ld->ac', Soovv, t1[ka], t1[kl])

        op_group = kqrts.stars_ops[i]
        ka_prim = kpts.k2opk[ka, op_group]
        mask = np.isin(ka_prim, ka_ibz_bz)
        for iop, ka_p in zip(op_group[mask], ka_prim[mask]):
            rmat_vv = rmat.vv[ka][iop]
            Fac[ka_p] += einsum('ac,ae,cf->ef', fock, rmat_vv.conj(), rmat_vv)
    return Fac

def Loo(kpts, kqrts, t1, t2, eris, rmat):
    nkpts, nocc, nvir = t1.shape
    fov = eris.fock[:,:nocc,nocc:]

    Lki = cc_Foo(kpts, kqrts, t1, t2, eris, rmat)
    for ki_ibz in range(kpts.nkpts_ibz):
        ki = kpts.ibz2bz[ki_ibz]
        Lki[ki] += einsum('kc,ic->ki', fov[ki], t1[ki])

    for i, kq in enumerate(kqrts.kqrts_ibz):
        ki, kl, ka, kb = kq
        if ki == ka:
            fock = (2*einsum('klic,lc->ki', eris.ooov[ki,kl,ki], t1[kl])
                     -einsum('lkic,lc->ki', eris.ooov[kl,ki,ki], t1[kl]))
            for _, iop in kqrts.loop_stabilizer(i):
                rmat_oo = rmat.oo[ki][iop]
                Lki[ki] += einsum('ki,km,in->mn', fock, rmat_oo.conj(), rmat_oo)
    return Lki

def Lvv(kpts, kqrts, t1, t2, eris, rmat):
    nkpts, nocc, nvir = t1.shape
    fov = eris.fock[:,:nocc,nocc:]

    Lac = cc_Fvv(kpts, kqrts, t1, t2, eris, rmat)
    for ka_ibz in range(kpts.nkpts_ibz):
        ka = kpts.ibz2bz[ka_ibz]
        Lac[ka] += -einsum('kc,ka->ac', fov[ka], t1[ka])

    for i, kq in enumerate(kqrts.kqrts_ibz):
        ka, kk, kc, kl = kq
        if ka == kc:
            Svovv = 2 * eris.vovv[ka,kk,ka] - eris.vovv[ka,kk,kk].transpose(0,1,3,2)
            fock = einsum('akcd,kd->ac', Svovv, t1[kk])
            for _, iop in kqrts.loop_stabilizer(i):
                rmat_vv = rmat.vv[ka][iop]
                Lac[ka] += einsum('ac,ae,cf->ef', fock, rmat_vv.conj(), rmat_vv)
    return Lac


def _s2_index(kqrts_ibz):
    kqrts_ibz_t = kqrts_ibz[:, np.array([1,0,3,2])]
    diff = kqrts_ibz_t[:,None,:] - kqrts_ibz[None,:,:]
    idx = np.sum(abs(diff), axis=-1)
    idx = np.array(np.where(idx == 0)).T
    idx1 = np.where(idx[:,0] < idx[:,1])[0]
    return idx[idx1][:,1]

def cc_Woooo(kpts, kqrts, t1, t2, eris, rmat, out=None):
    nkpts, nocc, nvir = t1.shape
    if out is None: #default incore
        #Wklij = ktensor.empty_like(eris.oooo)
        metadata = {'kpts': kpts, 'kqrts': kqrts, 'rmat': rmat,
                    'label': 'oooo', 'trans': 'ccnn',
                    'incore': True}
        Wklij = ktensor.empty([nocc,nocc,nocc,nocc], dtype=t1.dtype,
                              metadata=metadata)
    else:
        Wklij = out

    s2_idx = _s2_index(kqrts.kqrts_ibz)

    for i, kq in enumerate(kqrts.kqrts_ibz):
        if i in s2_idx:
            continue
        kk, kl, ki, kj = kq
        oooo  = einsum('klic,jc->klij',eris.ooov[kk,kl,ki],t1[kj])
        oooo += einsum('lkjc,ic->klij',eris.ooov[kl,kk,kj],t1[ki])
        oooo += eris.oooo[kk,kl,ki]

        vvoo = eris.oovv[kk,kl].transpose(0,3,4,1,2).reshape(nkpts*nvir,nvir,nocc,nocc)
        t2t  = t2[ki,kj].copy().transpose(0,3,4,1,2)
        t2t[ki] += einsum('ic,jd->cdij',t1[ki],t1[kj])
        t2t = t2t.reshape(nkpts*nvir,nvir,nocc,nocc)
        oooo += einsum('cdkl,cdij->klij',vvoo,t2t)
        Wklij[kk,kl,ki] = oooo

    for i, kq in enumerate(kqrts.kqrts_ibz[s2_idx]):
        kl, kk, kj, ki = kq
        Wklij[kl,kk,kj] = Wklij[kk,kl,ki].transpose(1,0,3,2)
    return Wklij

def cc_Wvvvv(kpts, kqrts, t1, t2, eris, rmat, out=None):
    nkpts, nocc, nvir = t1.shape
    if out is None: #default incore
        #Wabcd = ktensor.empty_like(eris.vvvv)
        metadata = {'kpts': kpts, 'kqrts': kqrts, 'rmat': rmat,
                    'label': 'vvvv', 'trans': 'ccnn',
                    'incore': True}
        Wabcd = ktensor.empty([nvir,nvir,nvir,nvir], dtype=t1.dtype,
                              metadata=metadata)
    else:
        Wabcd = out

    s2_idx = _s2_index(kqrts.kqrts_ibz)

    for i, kq in enumerate(kqrts.kqrts_ibz):
        if i in s2_idx:
            continue
        ka, kb, kc, kd = kq
        vvvv  = einsum('akcd,kb->abcd', eris.vovv[ka,kb,kc], -t1[kb])
        vvvv += einsum('bkdc,ka->abcd', eris.vovv[kb,ka,kd], -t1[ka])
        vvvv += eris.vvvv[ka,kb,kc]
        Wabcd[ka,kb,kc] = vvvv

    for i, kq in enumerate(kqrts.kqrts_ibz[s2_idx]):
        kb, ka, kd, kc = kq
        Wabcd[kb,ka,kd] = Wabcd[ka,kb,kc].transpose(1,0,3,2)
    return Wabcd

def cc_Wvoov(kpts, kqrts, t1, t2, eris, rmat, out=None):
    nkpts, nocc, nvir = t1.shape
    if out is None: #default incore
        #Wakic = ktensor.empty_like(eris.voov)
        metadata = {'kpts': kpts, 'kqrts': kqrts, 'rmat': rmat,
                    'label': 'voov', 'trans': 'ccnn',
                    'incore': True}
        Wakic = ktensor.empty([nvir,nocc,nocc,nvir], dtype=t1.dtype,
                              metadata=metadata)
    else:
        Wakic = out

    for i, kq in enumerate(kqrts.kqrts_ibz):
        ka, kk, ki, kc = kq
        voov  = einsum('akdc,id->akic', eris.vovv[ka,kk,ki], t1[ki])
        voov -= einsum('lkic,la->akic', eris.ooov[ka,kk,ki], t1[ka])
        voov += eris.voov[ka,kk,ki]

        kd = ki
        tau = t2[:,ki,ka].copy()
        tau[ka] += 2*einsum('id,la->liad', t1[kd], t1[ka])
        oovv_tmp = np.array(eris.oovv[kk,:,kc])
        voov -= 0.5*einsum('xklcd,xliad->akic', oovv_tmp, tau)
        Soovv_tmp = 2*oovv_tmp - eris.oovv[:,kk,kc].transpose(0,2,1,3,4)
        voov += 0.5*einsum('xklcd,xilad->akic', Soovv_tmp, t2[ki,:,ka])

        Wakic[ka,kk,ki] = voov
    return Wakic

def cc_Wvovo(kpts, kqrts, t1, t2, eris, rmat, out=None):
    nkpts, nocc, nvir = t1.shape
    if out is None: #default incore
        #Wakci = np.empty((nkpts,nkpts,nkpts,nvir,nocc,nvir,nocc), t1.dtype)
        metadata = {'kpts': kpts, 'kqrts': kqrts, 'rmat': rmat,
                    'label': 'vovo', 'trans': 'ccnn',
                    'incore': True}
        Wakci = ktensor.empty([nvir,nocc,nvir,nocc], dtype=t1.dtype,
                              metadata=metadata)
    else:
        Wakci = out

    for i, kq in enumerate(kqrts.kqrts_ibz):
        ka, kk, kc, ki = kq
        vovo  = einsum('akcd,id->akci',eris.vovv[ka,kk,kc],t1[ki])
        vovo -= einsum('klic,la->akci',eris.ooov[kk,ka,ki],t1[ka])
        vovo += np.asarray(eris.ovov[kk,ka,ki]).transpose(1,0,3,2)

        oovvf = eris.oovv[:,kk,kc].reshape(nkpts*nocc,nocc,nvir,nvir)
        t2f   = t2[:,ki,ka].copy()
        kd = ki
        t2f[ka] += 2*einsum('id,la->liad',t1[kd],t1[ka])
        t2f = t2f.reshape(nkpts*nocc,nocc,nvir,nvir)
        vovo -= 0.5*einsum('lkcd,liad->akci',oovvf,t2f)

        Wakci[ka,kk,kc] = vovo
    return Wakci



#!/usr/bin/env python
# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from pyscf import lib
from pyscf.lib import logger

#einsum = np.einsum
einsum = lib.einsum

# Ref: Gauss and Stanton, J. Chem. Phys. 103, 3561 (1995) Table III

# Section (a)

def make_tau(t2, t1a, t1b, fac=1, out=None):
    t1t1 = einsum('ia,jb->ijab', fac*0.5*t1a, t1b)
    t1t1 = t1t1 - t1t1.transpose(1,0,2,3)
    tau1 = t1t1 - t1t1.transpose(0,1,3,2)
    tau1 += t2
    return tau1

def cc_Fvv(t1, t2, eris):
    nocc, nvir = t1.shape
    fov = eris.fock[:nocc,nocc:]
    fvv = eris.fock[nocc:,nocc:]
    eris_vovv = np.asarray(eris.ovvv).transpose(1,0,3,2)
    tau_tilde = make_tau(t2, t1, t1,fac=0.5)
    Fae = fvv - 0.5*einsum('me,ma->ae',fov, t1)
    Fae += einsum('mf,amef->ae', t1, eris_vovv)
    Fae -= 0.5*einsum('mnaf,mnef->ae', tau_tilde, eris.oovv)
    return Fae

def cc_Foo(t1, t2, eris):
    nocc, nvir = t1.shape
    fov = eris.fock[:nocc,nocc:]
    foo = eris.fock[:nocc,:nocc]
    tau_tilde = make_tau(t2, t1, t1,fac=0.5)
    Fmi = ( foo + 0.5*einsum('me,ie->mi',fov, t1)
            + einsum('ne,mnie->mi', t1, eris.ooov)
            + 0.5*einsum('inef,mnef->mi', tau_tilde, eris.oovv) )
    return Fmi

def cc_Fov(t1, t2, eris):
    nocc, nvir = t1.shape
    fov = eris.fock[:nocc,nocc:]
    Fme = fov + einsum('nf,mnef->me', t1, eris.oovv)
    return Fme

def cc_Woooo(t1, t2, eris):
    tau = make_tau(t2, t1, t1)
    tmp = einsum('je,mnie->mnij', t1, eris.ooov)
    Wmnij = eris.oooo + tmp - tmp.transpose(0,1,3,2)
    Wmnij += 0.25*einsum('ijef,mnef->mnij', tau, eris.oovv)
    return Wmnij

def cc_Wvvvv(t1, t2, eris):
    tau = make_tau(t2, t1, t1)
    eris_ovvv = np.asarray(eris.ovvv)
    tmp = einsum('mb,mafe->bafe', t1, eris_ovvv)
    Wabef = np.asarray(eris.vvvv) - tmp + tmp.transpose(1,0,2,3)
    Wabef += einsum('mnab,mnef->abef', tau, 0.25*np.asarray(eris.oovv))
    return Wabef

def cc_Wovvo(t1, t2, eris):
    eris_ovvo = -np.asarray(eris.ovov).transpose(0,1,3,2)
    eris_oovo = -np.asarray(eris.ooov).transpose(0,1,3,2)
    Wmbej  = einsum('jf,mbef->mbej', t1, eris.ovvv)
    Wmbej -= einsum('nb,mnej->mbej', t1, eris_oovo)
    Wmbej -= 0.5*einsum('jnfb,mnef->mbej', t2, eris.oovv)
    Wmbej -= einsum('jf,nb,mnef->mbej', t1, t1, eris.oovv)
    Wmbej += eris_ovvo
    return Wmbej

### Section (b)

def Fvv(t1, t2, eris):
    ccFov = cc_Fov(t1, t2, eris)
    Fae = cc_Fvv(t1, t2, eris) - 0.5*einsum('ma,me->ae', t1,ccFov)
    return Fae

def Foo(t1, t2, eris):
    ccFov = cc_Fov(t1, t2, eris)
    Fmi = cc_Foo(t1, t2, eris) + 0.5*einsum('ie,me->mi', t1,ccFov)
    return Fmi

def Fov(t1, t2, eris):
    Fme = cc_Fov(t1, t2, eris)
    return Fme

def Woooo(t1, t2, eris):
    tau = make_tau(t2, t1, t1)
    Wmnij = 0.25*einsum('ijef,mnef->mnij', tau, eris.oovv)
    Wmnij += cc_Woooo(t1, t2, eris)
    return Wmnij

def Wvvvv(t1, t2, eris):
    tau = make_tau(t2, t1, t1)
    Wabef = cc_Wvvvv(t1, t2, eris)
    Wabef += einsum('mnab,mnef->abef', tau, .25*np.asarray(eris.oovv))
    return Wabef

def Wovvo(t1, t2, eris):
    Wmbej = -0.5*einsum('jnfb,mnef->mbej', t2, eris.oovv)
    Wmbej += cc_Wovvo(t1, t2, eris)
    return Wmbej

def Wooov(t1, t2, eris):
    Wmnie = einsum('if,mnfe->mnie', t1, eris.oovv)
    Wmnie += eris.ooov
    return Wmnie

def Wvovv(t1, t2, eris):
    Wamef = einsum('na,nmef->amef', -t1, eris.oovv)
    Wamef -= np.asarray(eris.ovvv).transpose(1,0,2,3)
    return Wamef

def Wovoo(t1, t2, eris):
    eris_ovvo = -np.asarray(eris.ovov).transpose(0,1,3,2)
    tmp1 = einsum('mnie,jnbe->mbij', eris.ooov, t2)
    tmp2 = einsum('ie,mbej->mbij', t1, eris_ovvo)
    tmp2 -= einsum('ie,njbf,mnef->mbij', t1, t2, eris.oovv)
    FFov = Fov(t1, t2, eris)
    WWoooo = Woooo(t1, t2, eris)
    tau = make_tau(t2, t1, t1)
    Wmbij = einsum('me,ijbe->mbij', -FFov, t2)
    Wmbij -= einsum('nb,mnij->mbij', t1, WWoooo)
    Wmbij += 0.5 * einsum('mbef,ijef->mbij', eris.ovvv, tau)
    Wmbij += tmp1 - tmp1.transpose(0,1,3,2)
    Wmbij += tmp2 - tmp2.transpose(0,1,3,2)
    Wmbij += np.asarray(eris.ooov).conj().transpose(2,3,0,1)
    return Wmbij

def Wvvvo(t1, t2, eris, _Wvvvv=None):
    eris_ovvo = -np.asarray(eris.ovov).transpose(0,1,3,2)
    eris_vvvo = -np.asarray(eris.ovvv).transpose(2,3,1,0).conj()
    eris_oovo = -np.asarray(eris.ooov).transpose(0,1,3,2)
    tmp1 = einsum('mbef,miaf->abei', eris.ovvv, t2)
    tmp2 = einsum('ma,mbei->abei', t1, eris_ovvo)
    tmp2 -= einsum('ma,nibf,mnef->abei', t1, t2, eris.oovv)
    FFov = Fov(t1, t2, eris)
    tau = make_tau(t2, t1, t1)
    Wabei  = 0.5 * einsum('mnei,mnab->abei', eris_oovo, tau)
    Wabei -= einsum('me,miab->abei', FFov, t2)
    Wabei += eris_vvvo
    Wabei -= tmp1 - tmp1.transpose(1,0,2,3)
    Wabei -= tmp2 - tmp2.transpose(1,0,2,3)
    nocc,nvir = t1.shape
    if _Wvvvv is None:
        _Wvvvv = Wvvvv(t1, t2, eris)
    Wabei += einsum('abef,if->abei', _Wvvvv, t1)
    return Wabei

########################################
# T3[2] related contributions to T1/T2
########################################

def get_t3p2_imds_slow(cc, t1, t2, eris=None, t3p2_ip_out=None, t3p2_ea_out=None):
    """Calculates T1, T2 amplitudes corrected by second-order T3 contribution
    and intermediates used in IP/EA-CCSD(T)a

    Args:
        cc (:obj:`GCCSD`):
            Object containing coupled-cluster results.
        t1 (:obj:`ndarray`):
            T1 amplitudes.
        t2 (:obj:`ndarray`):
            T2 amplitudes from which the T3[2] amplitudes are formed.
        eris (:obj:`_PhysicistsERIs`):
            Antisymmetrized electron-repulsion integrals in physicist's notation.
        t3p2_ip_out (:obj:`ndarray`):
            Store results of the intermediate used in IP-EOM-CCSD(T)a.
        t3p2_ea_out (:obj:`ndarray`):
            Store results of the intermediate used in EA-EOM-CCSD(T)a.

    Returns:
        delta_ccsd (float):
            Difference of perturbed and unperturbed CCSD ground-state energy,
                energy(T1 + T1[2], T2 + T2[2]) - energy(T1, T2)
        pt1 (:obj:`ndarray`):
            Perturbatively corrected T1 amplitudes.
        pt2 (:obj:`ndarray`):
            Perturbatively corrected T2 amplitudes.

    Reference:
        D. A. Matthews, J. F. Stanton "A new approach to approximate..."
            JCP 145, 124102 (2016); DOI:10.1063/1.4962910, Equation 14
        Shavitt and Bartlett "Many-body Methods in Physics and Chemistry"
            2009, Equation 10.33
    """
    if eris is None:
        eris = cc.ao2mo()
    fock = eris.fock
    nocc, nvir = t1.shape

    fov = fock[:nocc, nocc:]
    #foo = fock[:nocc, :nocc].diagonal()
    #fvv = fock[nocc:, nocc:].diagonal()

    mo_e_o = eris.mo_energy[:nocc]
    mo_e_v = eris.mo_energy[nocc:]

    oovv = np.asarray(eris.oovv)
    ovvv = np.asarray(eris.ovvv)
    ooov = np.asarray(eris.ooov)
    oovv = np.asarray(eris.oovv)
    vooo = np.asarray(ooov).conj().transpose(3, 2, 1, 0)
    vvvo = np.asarray(ovvv).conj().transpose(3, 2, 1, 0)

    ccsd_energy = cc.energy(t1, t2, eris)
    dtype = np.result_type(t1, t2)
    if np.issubdtype(dtype, np.dtype(complex).type):
        logger.error(cc, 't3p2 imds has not been strictly checked for use with complex integrals')

    if t3p2_ip_out is None:
        t3p2_ip_out = np.zeros((nocc,nvir,nocc,nocc), dtype=dtype)
    Wmcik = t3p2_ip_out

    if t3p2_ea_out is None:
        t3p2_ea_out = np.zeros((nvir,nvir,nvir,nocc), dtype=dtype)
    Wacek = t3p2_ea_out

    tmp_t3 = lib.einsum('bcdk,ijad->ijkabc', vvvo, t2)
    tmp_t3 -= lib.einsum('cmkj,imab->ijkabc', vooo, t2)

    # P(ijk)
    tmp_t3 = (tmp_t3 + tmp_t3.transpose(1, 2, 0, 3, 4, 5) +
                       tmp_t3.transpose(2, 0, 1, 3, 4, 5))
    # P(abc)
    tmp_t3 = (tmp_t3 + tmp_t3.transpose(0, 1, 2, 4, 5, 3) +
                       tmp_t3.transpose(0, 1, 2, 5, 3, 4))

    eia = mo_e_o[:, None] - mo_e_v[None, :]
    eijab = eia[:, None, :, None] + eia[None, :, None, :]
    eijkabc = eijab[:, :, None, :, :, None] + eia[None, None, :, None, None, :]
    tmp_t3 /= eijkabc

    pt1 = 0.25 * lib.einsum('mnef,imnaef->ia', oovv, tmp_t3)

    pt2 = lib.einsum('ijmabe,me->ijab', tmp_t3, fov)
    tmp2 = ovvv - 0.0*lib.einsum('nmef,nb->mbfe', oovv, t1)
    tmp = lib.einsum('ijmaef,mbfe->ijab', tmp_t3, tmp2)
    tmp = tmp - tmp.transpose(0, 1, 3, 2)  # P(ab)
    tmp *= 0.5
    pt2 = pt2 + tmp
    tmp2 = ooov + 0.0*lib.einsum('mnef,jf->nmje', oovv, t1)
    tmp = lib.einsum('inmabe,nmje->ijab', tmp_t3, tmp2)
    tmp *= -0.5
    tmp = tmp - tmp.transpose(1, 0, 2, 3)  # P(ij)
    pt2 = pt2 + tmp

    eia = mo_e_o[:, None] - mo_e_v[None, :]
    eijab = eia[:, None, :, None] + eia[None, :, None, :]

    pt1 /= eia
    pt2 /= eijab

    pt1 = pt1 + t1
    pt2 = pt2 + t2

    Wmcik += 0.5*lib.einsum('ijkabc,mjab->mcik', tmp_t3, oovv)
    Wacek += -0.5*lib.einsum('ijkabc,ijeb->acek', tmp_t3, oovv)

    delta_ccsd_energy = cc.energy(pt1, pt2, eris) - ccsd_energy
    logger.info(cc, 'CCSD energy T3[2] correction : %16.12e', delta_ccsd_energy)
    return delta_ccsd_energy, pt1, pt2, Wmcik, Wacek



#!/usr/bin/env python
# Copyright 2014-2023 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#
# Adapted for complex integral and general spin:
#         Huanchen Zhai <hczhai.ok@gmail.com>    Nov 16, 2022
#

import numpy
from pyscf import lib
from pyscf import ao2mo
from pyscf.fci import cistring
from pyscf.fci import direct_spin1
from pyscf.lib import logger


def contract_2e(eri, fcivec, norb, nelec, opt=None):
    '''Compute a^\\dagger_p a_q a^\\dagger_r a_s |CI>'''
    link_index = cistring.gen_linkstr_index(range(norb), nelec)
    na = cistring.num_strings(norb, nelec)
    t1 = numpy.zeros((norb, norb, na), dtype=eri.dtype)
    for str0, tab in enumerate(link_index):
        for a, i, str1, sign in tab:
            t1[a, i, str1] += sign * fcivec[str0]
    t1 = numpy.einsum('bjai,aiA->bjA', eri, t1, optimize=True)
    fcinew = numpy.zeros_like(fcivec)
    for str0, tab in enumerate(link_index):
        for a, i, str1, sign in tab:
            fcinew[str1] += sign * t1[a, i, str0]
    return fcinew


def absorb_h1e(h1e, eri, norb, nelec, fac=1):
    '''Modify 2e Hamiltonian to include 1e Hamiltonian contribution.'''
    f1e = h1e + 0.5 * numpy.einsum('jiik->jk', eri.transpose(0, 2, 3, 1), optimize=True)
    h2e = -1.0 * eri.transpose(0, 3, 2, 1)
    h2e[numpy.diag_indices(norb)] += f1e * (2.0 / (nelec + 1e-100))
    return h2e * fac


def make_hdiag(h1e, eri, norb, nelec, opt=None):
    occslist = cistring.gen_occslst(range(norb), nelec)
    diagjk = numpy.einsum('iijj->ij', eri.copy(), optimize=True)
    diagjk -= numpy.einsum('ijji->ij', eri, optimize=True)
    hdiag = []
    for occ in occslist:
        e1 = h1e[occ, occ].sum()
        e2 = diagjk[occ][:, occ].sum()
        hdiag.append(e1 + e2 * 0.5)
    return numpy.array(hdiag)


def kernel(h1e, eri, norb, nelec, ecore=0, nroots=1, verbose=3):
    h2e = absorb_h1e(h1e, eri, norb, nelec, 0.5)
    hop = lambda c: contract_2e(h2e, c, norb, nelec)
    hdiag = make_hdiag(h1e, eri, norb, nelec)
    precond = lambda x, e, *args: x / (hdiag - e + 1e-4)
    ci0 = get_init_guess(norb, nelec, nroots, hdiag)
    e, c = lib.davidson(hop, ci0, precond, verbose=verbose, nroots=nroots)
    return e + ecore, c


# dm_pq = <|p^+ q|>
def make_rdm1_slow(fcivec, norb, nelec, link_index=None):
    if link_index is None:
        link_index = cistring.gen_linkstr_index(range(norb), nelec)
    rdm1 = numpy.zeros((norb, norb), dtype=fcivec.dtype)
    for str0, tab in enumerate(link_index):
        for a, i, str1, sign in tab:
            rdm1[a, i] += sign * numpy.dot(fcivec[str1].conj(), fcivec[str0])
    return rdm1


# dm_pq = <|p^+ q|>
def make_rdm1(fcivec, norb, nelec, link_index=None):
    if nelec == 0:
        return numpy.zeros((norb, norb), dtype=fcivec.dtype)
    nd = cistring.num_strings(norb, nelec - 1)
    index_d = cistring.gen_des_str_index(range(norb), nelec)
    dci = numpy.zeros((nd, norb), dtype=fcivec.dtype)
    for str0, tab in enumerate(index_d):
        for _, i, str1, sign in tab:
            dci[str1, i] += sign * fcivec[str0]
    return dci.conj().T @ dci


# dm_pq,rs = <|p^+ q r^+ s|>
def make_rdm12(fcivec, norb, nelec, link_index=None, reorder=True):
    if link_index is None:
        link_index = cistring.gen_linkstr_index(range(norb), nelec)
    link_index = cistring.gen_linkstr_index(range(norb), nelec)
    rdm1 = numpy.zeros((norb, norb), dtype=fcivec.dtype)
    rdm2 = numpy.zeros((norb, norb, norb, norb), dtype=fcivec.dtype)
    for str0, tab in enumerate(link_index):
        t1 = numpy.zeros((norb, norb), dtype=fcivec.dtype)
        for a, i, str1, sign in tab:
            t1[i, a] += sign * fcivec[str1]
        rdm1 += fcivec[str0].conj() * t1
        # i^+ j|0> => <0|j^+ i, so swap i and j
        rdm2 += numpy.einsum('ij,kl->jikl', t1.conj(), t1, optimize=True)
    return reorder_rdm(rdm1, rdm2) if reorder else (rdm1, rdm2)


def reorder_rdm(rdm1, rdm2, inplace=True):
    nmo = rdm1.shape[0]
    rdm2 = (rdm2 if inplace else rdm2.copy()).reshape(nmo, nmo, nmo, nmo)
    rdm2[(slice(None), ) + numpy.diag_indices(nmo)] -= rdm1[:, None, :]
    rdm2 = -rdm2.transpose(0, 3, 2, 1)
    return rdm1, rdm2


def get_init_guess(norb, nelec, nroots, hdiag):
    '''Initial guess is the single Slater determinant
    '''
    na = cistring.num_strings(norb, nelec)
    ci0 = []
    if hdiag.size <= nroots:
        addrs = numpy.arange(hdiag.size)
    else:
        addrs = numpy.argpartition(hdiag, nroots-1)[:nroots]
    for addr in addrs:
        x = numpy.zeros((na, ), dtype=hdiag.dtype)
        x[addr] = 1
        ci0.append(x)

    # Add noise
    ci0[0][0 ] += 1e-5 + 1e-6j if ci0[0].dtype == complex else 1e-5
    ci0[0][-1] -= 1e-5 + 1e-6j if ci0[0].dtype == complex else 1e-5
    return ci0


def kernel_dhf(fci, h1e, eri, norb, nelec, ci0=None, link_index=None,
               tol=None, lindep=None, max_cycle=None, max_space=None,
               nroots=None, davidson_only=None, pspace_size=None,
               max_memory=None, verbose=None, ecore=0, **kwargs):
    if nroots is None: nroots = fci.nroots
    if davidson_only is None: davidson_only = fci.davidson_only
    if pspace_size is None: pspace_size = fci.pspace_size
    if max_memory is None:
        max_memory = fci.max_memory - lib.current_memory()[0]
    log = logger.new_logger(fci, verbose)
    link_index = cistring.gen_linkstr_index(range(norb), nelec)
    na = cistring.num_strings(norb, nelec)
    hdiag = fci.make_hdiag(h1e, eri, norb, nelec)
    nroots = min(hdiag.size, nroots)
    precond = lib.make_diag_precond(hdiag, fci.level_shift)

    h2e = fci.absorb_h1e(h1e, eri, norb, nelec, 0.5)
    def hop(c):
        return fci.contract_2e(h2e, c, norb, nelec, link_index)

    def init_guess():
        if callable(getattr(fci, 'get_init_guess', None)):
            return fci.get_init_guess(norb, nelec, nroots, hdiag)
        else:
            x0 = []
            for i in range(nroots):
                x = numpy.zeros(na, dtype=eri.dtype)
                x[i] = 1
                x0.append(x)
            return x0

    if ci0 is None:
        ci0 = init_guess  # lazy initialization to reduce memory footprint
    elif not callable(ci0):
        if len(ci0) < nroots:
            for i in range(len(ci0), nroots):
                x = numpy.zeros(na, dtype=eri.dtype)
                x[i] = 1
                ci0.append(x)

    if tol is None: tol = fci.conv_tol
    if lindep is None: lindep = fci.lindep
    if max_cycle is None: max_cycle = fci.max_cycle
    if max_space is None: max_space = fci.max_space
    tol_residual = getattr(fci, 'conv_tol_residual', None)

    with lib.with_omp_threads(fci.threads):
        e, c = fci.eig(hop, ci0, precond, tol=tol, lindep=lindep,
                       max_cycle=max_cycle, max_space=max_space, nroots=nroots,
                       max_memory=max_memory, verbose=log, follow_state=True,
                       tol_residual=tol_residual, **kwargs)
    if nroots > 1:
        return e + ecore, [ci.view(direct_spin1.FCIvector) for ci in c]
    else:
        return e + ecore, c.view(direct_spin1.FCIvector)


###############################################################
# ghf/dhf-integral direct-CI driver
###############################################################

class FCISolver(direct_spin1.FCISolver):

    def absorb_h1e(self, h1e, eri, norb, nelec, fac=1):
        return absorb_h1e(h1e, eri, norb, nelec, fac)

    def make_hdiag(self, h1e, eri, norb, nelec):
        return make_hdiag(h1e, eri, norb, nelec)

    def pspace(self, h1e, eri, norb, nelec, hdiag, np=400):
        return NotImplemented

    def contract_1e(self, f1e, fcivec, norb, nelec, link_index=None, **kwargs):
        return NotImplemented

    def contract_2e(self, eri, fcivec, norb, nelec, link_index=None, **kwargs):
        return contract_2e(eri, fcivec, norb, nelec, link_index, **kwargs)

    def spin_square(self, fcivec, norb, nelec):
        return NotImplemented

    def make_rdm1(self, fcivec, norb, nelec, link_index=None):
        return make_rdm1(fcivec, norb, nelec, link_index)

    def make_rdm12(self, fcivec, norb, nelec, link_index=None, reorder=True):
        return make_rdm12(fcivec, norb, nelec, link_index, reorder)

    def make_rdm2(self, fcivec, norb, nelec, link_index=None, reorder=True):
        return self.make_rdm12(fcivec, norb, nelec, link_index, reorder)[1]

    def energy(self, h1e, eri, fcivec, norb, nelec, link_index=None):
        h2e = self.absorb_h1e(h1e, eri, norb, nelec, 0.5)
        ci1 = self.contract_2e(h2e, fcivec, norb, nelec, link_index)
        return numpy.dot(fcivec.conj(), ci1)

    def get_init_guess(self, norb, nelec, nroots, hdiag):
        return get_init_guess(norb, nelec, nroots, hdiag)

    def kernel(self, h1e, eri, norb, nelec, ci0=None,
               tol=None, lindep=None, max_cycle=None, max_space=None,
               nroots=None, davidson_only=None, pspace_size=None,
               orbsym=None, wfnsym=None, ecore=0, **kwargs):
        self.norb = norb
        self.nelec = nelec
        self.eci, self.ci = \
                kernel_dhf(self, h1e, eri, norb, nelec, ci0, None,
                           tol, lindep, max_cycle, max_space, nroots,
                           davidson_only, pspace_size, ecore=ecore, **kwargs)
        return self.eci, self.ci

FCI = FCISolver

if __name__ == '__main__':
    from functools import reduce
    from pyscf import gto
    from pyscf import scf

    mol = gto.Mole()
    mol.verbose = 0
    mol.output = None
    mol.atom = [
        ['H', ( 1.,-1.    , 0.   )],
        ['H', ( 0.,-1.    ,-1.   )],
        ['H', ( 1.,-0.5   ,-1.   )],
        ['H', ( 0.,-0.    ,-1.   )],
        ['H', ( 1.,-0.5   , 0.   )],
        ['H', ( 0., 1.    , 1.   )],
    ]
    mol.basis = 'sto-3g'
    mol.build()

    m = scf.GHF(mol).run(conv_tol=1e-14)

    norb = m.mo_coeff.shape[1]
    h1e = reduce(numpy.dot, (m.mo_coeff.conj().T, m.get_hcore(), m.mo_coeff))
    mo_a, mo_b = m.mo_coeff[:mol.nao], m.mo_coeff[mol.nao:]
    eri = ao2mo.restore(4, ao2mo.general(m._eri, (mo_a, mo_a, mo_b, mo_b)), norb)
    eri = eri + eri.transpose(1, 0)
    eri += ao2mo.restore(4, ao2mo.full(m._eri, mo_a), norb)
    eri += ao2mo.restore(4, ao2mo.full(m._eri, mo_b), norb)
    eri = ao2mo.restore(1, eri, norb)

    nelec = mol.nelectron - 2

    e1 = kernel(h1e, eri, norb, nelec)[0]
    print(e1, e1 - -7.9766331504361414)



#!/usr/bin/env python
# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''
Dirac Hartree-Fock
'''


from functools import reduce
import ctypes
import numpy
from pyscf import lib
from pyscf import gto
from pyscf.lib import logger
from pyscf.scf import hf
from pyscf.scf import _vhf
from pyscf.scf import chkfile
from pyscf.data import nist
from pyscf import __config__

zquatev = None
if getattr(__config__, 'scf_dhf_SCF_zquatev', True):
    try:
        # Install zquatev with
        # pip install git+https://github.com/sunqm/zquatev
        import zquatev
    except ImportError:
        pass

DEBUG = False


def kernel(mf, conv_tol=1e-9, conv_tol_grad=None,
           dump_chk=True, dm0=None, callback=None, conv_check=True):
    '''the modified SCF kernel for Dirac-Hartree-Fock.  In this kernel, the
    SCF is carried out in three steps.  First the 2-electron part is
    approximated by large component integrals (LL|LL); Next, (SS|LL) the
    interaction between large and small components are added; Finally,
    converge the SCF with the small component contributions (SS|SS)
    '''
    if conv_tol_grad is None:
        conv_tol_grad = numpy.sqrt(conv_tol)
        logger.info(mf, 'Set gradient conv threshold to %g', conv_tol_grad)
    if dm0 is None:
        dm = mf.get_init_guess(mf.mol, mf.init_guess)
    else:
        dm = dm0

    if mf.init_guess != 'chkfile':
        mf._coulomb_level = 'LLLL'
    cycles = 0
    if dm0 is None and mf._coulomb_level.upper() == 'LLLL':
        scf_conv, e_tot, mo_energy, mo_coeff, mo_occ \
                = hf.kernel(mf, 1e-2, 1e-1,
                            dump_chk, dm0=dm, callback=callback,
                            conv_check=False)
        cycles += mf.cycles
        dm = mf.make_rdm1(mo_coeff, mo_occ)
        mf._coulomb_level = 'SSLL'

    if mf.with_ssss:
        if dm0 is None and (mf._coulomb_level.upper() == 'SSLL' or
                            mf._coulomb_level.upper() == 'LLSS'):
            scf_conv, e_tot, mo_energy, mo_coeff, mo_occ \
                    = hf.kernel(mf, 1e-3, 1e-1,
                                dump_chk, dm0=dm, callback=callback,
                                conv_check=False)
            cycles += mf.cycles
            dm = mf.make_rdm1(mo_coeff, mo_occ)
        mf._coulomb_level = 'SSSS'
    else:
        mf._coulomb_level = 'SSLL'

    out = hf.kernel(mf, conv_tol, conv_tol_grad, dump_chk, dm0=dm,
                    callback=callback, conv_check=conv_check)
    mf.cycles = cycles + mf.cycles
    return out

def energy_elec(mf, dm=None, h1e=None, vhf=None):
    r'''Electronic part of Dirac-Hartree-Fock energy

    Args:
        mf : an instance of SCF class

    Kwargs:
        dm : 2D ndarray
            one-particle density matrix
        h1e : 2D ndarray
            Core hamiltonian
        vhf : 2D ndarray
            HF potential

    Returns:
        Hartree-Fock electronic energy and the Coulomb energy
    '''
    if dm is None: dm = mf.make_rdm1()
    if h1e is None: h1e = mf.get_hcore()
    if vhf is None: vhf = mf.get_veff(mf.mol, dm)
    e1 = numpy.einsum('ij,ji->', h1e, dm).real
    e_coul = numpy.einsum('ij,ji->', vhf, dm).real * .5
    logger.debug(mf, 'E1 = %.14g  E_coul = %.14g', e1, e_coul)

    if not mf.with_ssss and mf.ssss_approx == 'Visscher':
        e_coul += _visscher_ssss_correction(mf, dm)

    mf.scf_summary['e1'] = e1
    mf.scf_summary['e2'] = e_coul
    return e1+e_coul, e_coul

def _visscher_ssss_correction(mf, dm):
    '''
    Visscher point charge corrections for small component, TCA, 98, 68
    Note there is a small difference to Visscher's work. The model
    charges in Visscher's work are obtained from atomic calculations.
    Charges here are Mulliken charges on small components.
    '''
    aoslice = mf.mol.aoslice_2c_by_atom()
    n2c = dm[0].shape[0] // 2
    s = mf.get_ovlp()
    ss_mul_charges = []
    for p0, p1 in aoslice[:,2:] + n2c:
        mul_charge = numpy.einsum('ij,ji->', s[n2c:,p0:p1], dm[p0:p1,n2c:])
        ss_mul_charges.append(mul_charge.real)
    ss_mul_charges = numpy.array(ss_mul_charges)
    e_coul_ss = gto.classical_coulomb_energy(mf.mol, ss_mul_charges)
    mf.scf_summary['e_coul_ss'] = e_coul_ss
    logger.debug(mf, 'Visscher corrections for small component = %.14g', e_coul_ss)
    return e_coul_ss

def get_jk_coulomb(mol, dm, hermi=1, coulomb_allow='SSSS',
                   opt_llll=None, opt_ssll=None, opt_ssss=None,
                   omega=None, verbose=None):
    log = logger.new_logger(mol, verbose)
    t0 = (logger.process_clock(), logger.perf_counter())

    if hermi == 0 and DEBUG:
        # J matrix is symmetrized in this function which is only true for
        # density matrix with time reversal symmetry
        _ensure_time_reversal_symmetry(mol, dm)

    with mol.with_range_coulomb(omega):
        if coulomb_allow.upper() == 'LLLL':
            log.debug('Coulomb integral: (LL|LL)')
            j1, k1 = _call_veff_llll(mol, dm, hermi, opt_llll)
            log.timer_debug1('LLLL', *t0)
            n2c = j1.shape[1]
            vj = numpy.zeros_like(dm)
            vk = numpy.zeros_like(dm)
            vj[..., :n2c, :n2c] = j1
            vk[..., :n2c, :n2c] = k1
        elif coulomb_allow.upper() == 'SSLL' \
          or coulomb_allow.upper() == 'LLSS':
            log.debug('Coulomb integral: (LL|LL) + (SS|LL)')
            vj, vk = _call_veff_ssll(mol, dm, hermi, opt_ssll)
            t0 = log.timer_debug1('SSLL', *t0)
            j1, k1 = _call_veff_llll(mol, dm, hermi, opt_llll)
            log.timer_debug1('LLLL', *t0)
            n2c = j1.shape[1]
            vj[..., :n2c, :n2c] += j1
            vk[..., :n2c, :n2c] += k1
        else:  # coulomb_allow == 'SSSS'
            log.debug('Coulomb integral: (LL|LL) + (SS|LL) + (SS|SS)')
            vj, vk = _call_veff_ssll(mol, dm, hermi, opt_ssll)
            t0 = log.timer_debug1('SSLL', *t0)
            j1, k1 = _call_veff_llll(mol, dm, hermi, opt_llll)
            t0 = log.timer_debug1('LLLL', *t0)
            n2c = j1.shape[1]
            vj[..., :n2c, :n2c] += j1
            vk[..., :n2c, :n2c] += k1
            j1, k1 = _call_veff_ssss(mol, dm, hermi, opt_ssss)
            log.timer_debug1('SSSS', *t0)
            vj[..., n2c:, n2c:] += j1
            vk[..., n2c:, n2c:] += k1

    return vj, vk
get_jk = get_jk_coulomb


def get_hcore(mol):
    n2c = mol.nao_2c()
    n4c = n2c * 2
    c = lib.param.LIGHT_SPEED

    t  = mol.intor_symmetric('int1e_spsp_spinor') * .5
    vn = mol.intor_symmetric('int1e_nuc_spinor')
    wn = mol.intor_symmetric('int1e_spnucsp_spinor')
    h1e = numpy.empty((n4c, n4c), numpy.complex128)
    h1e[:n2c,:n2c] = vn
    h1e[n2c:,:n2c] = t
    h1e[:n2c,n2c:] = t
    h1e[n2c:,n2c:] = wn * (.25/c**2) - t
    return h1e

def get_ovlp(mol):
    n2c = mol.nao_2c()
    n4c = n2c * 2
    c = lib.param.LIGHT_SPEED

    s = mol.intor_symmetric('int1e_ovlp_spinor')
    t = mol.intor_symmetric('int1e_spsp_spinor')
    s1e = numpy.zeros((n4c, n4c), numpy.complex128)
    s1e[:n2c,:n2c] = s
    s1e[n2c:,n2c:] = t * (.5/c)**2
    return s1e

make_rdm1 = hf.make_rdm1

def init_guess_by_minao(mol):
    '''Initial guess in terms of the overlap to minimal basis.'''
    dm = hf.init_guess_by_minao(mol)
    return _proj_dmll(mol, dm, mol)

def init_guess_by_1e(mol):
    '''Initial guess from one electron system.'''
    return UHF(mol).init_guess_by_1e(mol)

def init_guess_by_atom(mol):
    '''Initial guess from atom calculation.'''
    dm = hf.init_guess_by_atom(mol)
    return _proj_dmll(mol, dm, mol)

def init_guess_by_huckel(mol):
    '''Initial guess from on-the-fly Huckel, doi:10.1021/acs.jctc.8b01089.'''
    dm = hf.init_guess_by_huckel(mol)
    return _proj_dmll(mol, dm, mol)

def init_guess_by_mod_huckel(mol):
    '''Initial guess from on-the-fly Huckel, doi:10.1021/acs.jctc.8b01089,
    employing the updated GWH rule from doi:10.1021/ja00480a005.'''
    dm = hf.init_guess_by_mod_huckel(mol)
    return _proj_dmll(mol, dm, mol)

def init_guess_by_sap(mol, mf):
    '''Generate initial guess density matrix from a superposition of
    atomic potentials (SAP), doi:10.1021/acs.jctc.8b01089.
    This is the Gaussian fit implementation, see doi:10.1063/5.0004046.

    Args:
        mol : MoleBase object
            the molecule object for which the initial guess is evaluated
        sap_basis : dict
            SAP basis in internal format (python dictionary)

    Returns:
        dm0 : ndarray
            SAP initial guess density matrix
    '''
    dm = hf.SCF.init_guess_by_sap(mf, mol)
    return _proj_dmll(mol, dm, mol)

def init_guess_by_chkfile(mol, chkfile_name, project=None):
    '''Read SCF chkfile and make the density matrix for 4C-DHF initial guess.

    Kwargs:
        project : None or bool
            Whether to project chkfile's orbitals to the new basis.  Note when
            the geometry of the chkfile and the given molecule are very
            different, this projection can produce very poor initial guess.
            In PES scanning, it is recommended to switch off project.

            If project is set to None, the projection is only applied when the
            basis sets of the chkfile's molecule are different to the basis
            sets of the given molecule (regardless whether the geometry of
            the two molecules are different).  Note the basis sets are
            considered to be different if the two molecules are derived from
            the same molecule with different ordering of atoms.
    '''
    from pyscf.scf import addons
    chk_mol, scf_rec = chkfile.load_scf(chkfile_name)
    if project is None:
        project = not gto.same_basis_set(chk_mol, mol)

    # Check whether the two molecules are similar
    if abs(mol.inertia_moment() - chk_mol.inertia_moment()).sum() > 0.5:
        logger.warn(mol, "Large deviations found between the input "
                    "molecule and the molecule from chkfile\n"
                    "Initial guess density matrix may have large error.")

    if project:
        s = get_ovlp(mol)

    def fproj(mo):
        #TODO: check if mo is GHF orbital
        if project:
            mo = addons.project_mo_r2r(chk_mol, mo, mol)
            norm = numpy.einsum('pi,pi->i', mo.conj(), s.dot(mo))
            mo /= numpy.sqrt(norm)
        return mo

    mo = scf_rec['mo_coeff']
    mo_occ = scf_rec['mo_occ']
    if numpy.iscomplexobj(mo[0]):  # DHF
        dm = make_rdm1(fproj(mo), mo_occ)
    else:
        if mo[0].ndim == 1: # nr-RHF
            dm = reduce(numpy.dot, (mo*mo_occ, mo.T))
        else: # nr-UHF
            dm = (reduce(numpy.dot, (mo[0]*mo_occ[0], mo[0].T)) +
                  reduce(numpy.dot, (mo[1]*mo_occ[1], mo[1].T)))
        dm = _proj_dmll(chk_mol, dm, mol)
    return dm


def get_init_guess(mol, key='minao', **kwargs):
    '''Generate density matrix for initial guess

    Kwargs:
        key : str
            One of 'minao', 'atom', 'huckel', 'mod_huckel', 'hcore', '1e', 'sap', 'chkfile'.
    '''
    return UHF(mol).get_init_guess(mol, key, **kwargs)

def time_reversal_matrix(mol, mat):
    ''' T(A_ij) = A[T(i),T(j)]^*
    '''
    tao = numpy.asarray(mol.time_reversal_map())
    n2c = tao.size
    # tao(i) = -j  means  T(f_i) = -f_j
    # tao(i) =  j  means  T(f_i) =  f_j
    idx = abs(tao) - 1  # -1 for C indexing convention
    #:signL = [(1 if x>0 else -1) for x in tao]
    #:sign = numpy.hstack((signL, signL))

    #:tmat = numpy.empty_like(mat)
    #:for j in range(mat.__len__()):
    #:    for i in range(mat.__len__()):
    #:        tmat[idx[i],idx[j]] = mat[j,i] * sign[i]*sign[j]
    #:return tmat.conjugate()
    sign_mask = tao < 0
    if mat.shape[0] == n2c * 2:
        idx = numpy.hstack((idx, idx+n2c))
        sign_mask = numpy.hstack((sign_mask, sign_mask))

    tmat = mat[idx[:,None], idx]
    tmat[sign_mask,:] *= -1
    tmat[:,sign_mask] *= -1
    return tmat.conj()

def analyze(mf, verbose=logger.DEBUG, **kwargs):
    from pyscf.tools import dump_mat
    log = logger.new_logger(mf, verbose)
    mo_energy = mf.mo_energy
    mo_occ = mf.mo_occ
    mo_coeff = mf.mo_coeff

    mf.dump_scf_summary(log)
    log.info('**** MO energy ****')
    for i in range(len(mo_energy)):
        if mo_occ[i] > 0:
            log.info('occupied MO #%d energy= %.15g occ= %g',
                     i+1, mo_energy[i], mo_occ[i])
        else:
            log.info('virtual MO #%d energy= %.15g occ= %g',
                     i+1, mo_energy[i], mo_occ[i])
    mol = mf.mol
    if mf.verbose >= logger.DEBUG1:
        log.debug(' ** MO coefficients of large component of positive state (real part) **')
        label = mol.spinor_labels()
        n2c = mo_coeff.shape[0] // 2
        dump_mat.dump_rec(mf.stdout, mo_coeff[n2c:,:n2c].real, label, start=1)
    dm = mf.make_rdm1(mo_coeff, mo_occ)
    pop_chg = mf.mulliken_pop(mol, dm, mf.get_ovlp(), log)
    dip = mf.dip_moment(mol, dm, verbose=log)
    return pop_chg, dip

def mulliken_pop(mol, dm, s=None, verbose=logger.DEBUG):
    r'''Mulliken population analysis

    .. math:: M_{ij} = D_{ij} S_{ji}

    Mulliken charges

    .. math:: \delta_i = \sum_j M_{ij}

    '''
    if s is None: s = get_ovlp(mol)
    log = logger.new_logger(mol, verbose)
    pop = numpy.einsum('ij,ji->i', dm, s).real

    log.info(' ** Mulliken pop  **')
    for i, s in enumerate(mol.spinor_labels()):
        log.info('pop of  %s %10.5f', s, pop[i])

    log.note(' ** Mulliken atomic charges  **')
    chg = numpy.zeros(mol.natm)
    for i, s in enumerate(mol.spinor_labels(fmt=None)):
        chg[s[0]] += pop[i]
    chg = mol.atom_charges() - chg
    for ia in range(mol.natm):
        symb = mol.atom_symbol(ia)
        log.note('charge of  %d%s =   %10.5f', ia, symb, chg[ia])
    return pop, chg

def dip_moment(mol, dm, unit='Debye', verbose=logger.NOTE, **kwargs):
    r''' Dipole moment calculation

    .. math::

        \mu_x = -\sum_{\mu}\sum_{\nu} P_{\mu\nu}(\nu|x|\mu) + \sum_A Q_A X_A\\
        \mu_y = -\sum_{\mu}\sum_{\nu} P_{\mu\nu}(\nu|y|\mu) + \sum_A Q_A Y_A\\
        \mu_z = -\sum_{\mu}\sum_{\nu} P_{\mu\nu}(\nu|z|\mu) + \sum_A Q_A Z_A

    where :math:`\mu_x, \mu_y, \mu_z` are the x, y and z components of dipole
    moment

    Args:
         mol: an instance of :class:`Mole`
         dm : a 2D ndarrays density matrices

    Return:
        A list: the dipole moment on x, y and z component
    '''
    log = logger.new_logger(mol, verbose)

    charges = mol.atom_charges()
    coords  = mol.atom_coords()
    charge_center = numpy.einsum('i,ix->x', charges, coords) / sum(charges)
    with mol.with_common_orig(charge_center):
        ll_dip = mol.intor_symmetric('int1e_r_spinor', comp=3)
        ss_dip = mol.intor_symmetric('int1e_sprsp_spinor', comp=3)

    n2c = mol.nao_2c()
    c = lib.param.LIGHT_SPEED
    dip = numpy.einsum('xij,ji->x', ll_dip, dm[:n2c,:n2c]).real
    dip+= numpy.einsum('xij,ji->x', ss_dip, dm[n2c:,n2c:]).real * (.5/c)**2

    dip *= -1.

    if unit.upper() == 'DEBYE':
        dip *= nist.AU2DEBYE
        log.note('Dipole moment(X, Y, Z, Debye): %8.5f, %8.5f, %8.5f', *dip)
    else:
        log.note('Dipole moment(X, Y, Z, A.U.): %8.5f, %8.5f, %8.5f', *dip)
    return dip

def get_grad(mo_coeff, mo_occ, fock_ao):
    '''DHF Gradients'''
    occidx = mo_occ > 0
    viridx = ~occidx
    g = reduce(numpy.dot, (mo_coeff[:,viridx].T.conj(), fock_ao,
                           mo_coeff[:,occidx]))
    return g.ravel()


# Kramers unrestricted Dirac-Hartree-Fock
class DHF(hf.SCF):
    __doc__ = hf.SCF.__doc__ + '''
    Attributes for Dirac-Hartree-Fock
        with_ssss : bool or string, for Dirac-Hartree-Fock only
            If False, ignore small component integrals (SS|SS).  Default is True.
        with_gaunt : bool, for Dirac-Hartree-Fock only
            Default is False.
        with_breit : bool, for Dirac-Hartree-Fock only
            Gaunt + gauge term.  Default is False.

    Examples:

    >>> mol = gto.M(atom='H 0 0 0; H 0 0 1', basis='ccpvdz', verbose=0)
    >>> mf = scf.RHF(mol)
    >>> e0 = mf.scf()
    >>> mf = scf.DHF(mol)
    >>> e1 = mf.scf()
    >>> print('Relativistic effects = %.12f' % (e1-e0))
    Relativistic effects = -0.000008854205
    '''

    conv_tol = getattr(__config__, 'scf_dhf_SCF_conv_tol', 1e-8)
    with_ssss = getattr(__config__, 'scf_dhf_SCF_with_ssss', True)
    with_gaunt = getattr(__config__, 'scf_dhf_SCF_with_gaunt', False)
    with_breit = getattr(__config__, 'scf_dhf_SCF_with_breit', False)
    # corrections for small component when with_ssss is set to False
    ssss_approx = getattr(__config__, 'scf_dhf_SCF_ssss_approx', 'Visscher')
    screening = True

    _keys = {'conv_tol', 'with_ssss', 'with_gaunt', 'with_breit', 'ssss_approx'}

    def __init__(self, mol):
        hf.SCF.__init__(self, mol)
        self._coulomb_level = 'SSSS' # 'SSSS' ~ LLLL+LLSS+SSSS

    def dump_flags(self, verbose=None):
        hf.SCF.dump_flags(self, verbose)
        log = logger.new_logger(self, verbose)
        log.info('with_ssss %s, with_gaunt %s, with_breit %s',
                 self.with_ssss, self.with_gaunt, self.with_breit)
        if not self.with_ssss:
            log.info('ssss_approx: %s', self.ssss_approx)
        log.info('light speed = %s', lib.param.LIGHT_SPEED)
        return self

    @lib.with_doc(get_hcore.__doc__)
    def get_hcore(self, mol=None):
        if mol is None:
            mol = self.mol
        return get_hcore(mol)

    @lib.with_doc(get_ovlp.__doc__)
    def get_ovlp(self, mol=None):
        if mol is None:
            mol = self.mol
        return get_ovlp(mol)

    def get_grad(self, mo_coeff, mo_occ, fock=None):
        if fock is None:
            dm1 = self.make_rdm1(mo_coeff, mo_occ)
            fock = self.get_hcore(self.mol) + self.get_veff(self.mol, dm1)
        return get_grad(mo_coeff, mo_occ, fock)

    def init_guess_by_minao(self, mol=None):
        '''Initial guess in terms of the overlap to minimal basis.'''
        if mol is None: mol = self.mol
        return init_guess_by_minao(mol)

    def init_guess_by_atom(self, mol=None):
        if mol is None: mol = self.mol
        return init_guess_by_atom(mol)

    @lib.with_doc(hf.SCF.init_guess_by_huckel.__doc__)
    def init_guess_by_huckel(self, mol=None):
        if mol is None: mol = self.mol
        logger.info(self, 'Initial guess from on-the-fly Huckel, doi:10.1021/acs.jctc.8b01089.')
        return init_guess_by_huckel(mol)

    @lib.with_doc(hf.SCF.init_guess_by_mod_huckel.__doc__)
    def init_guess_by_mod_huckel(self, mol=None):
        if mol is None: mol = self.mol
        logger.info(self, '''Initial guess from on-the-fly Huckel, doi:10.1021/acs.jctc.8b01089,
employing the updated GWH rule from doi:10.1021/ja00480a005.''')
        return init_guess_by_mod_huckel(mol)

    @lib.with_doc(hf.SCF.init_guess_by_sap.__doc__)
    def init_guess_by_sap(self, mol=None, **kwargs):
        if mol is None: mol = self.mol
        return init_guess_by_sap(mol, self)

    def init_guess_by_chkfile(self, chkfile=None, project=None):
        if chkfile is None: chkfile = self.chkfile
        return init_guess_by_chkfile(self.mol, chkfile, project=project)

    def build(self, mol=None):
        if self.verbose >= logger.WARN:
            self.check_sanity()
        return self

    def get_occ(self, mo_energy=None, mo_coeff=None):
        if mo_energy is None: mo_energy = self.mo_energy
        mol = self.mol
        c = lib.param.LIGHT_SPEED
        n4c = len(mo_energy)
        n2c = n4c // 2
        mo_occ = numpy.zeros(n2c * 2)
        nocc = mol.nelectron
        if mo_energy[n2c] > -1.999 * c**2:
            mo_occ[n2c:n2c+nocc] = 1
        else:
            logger.warn(self, 'Variational collapse. PES mo_energy %g < -2c^2',
                        mo_energy[n2c])
            lumo = mo_energy[mo_energy > -1.999 * c**2][nocc]
            mo_occ[mo_energy > -1.999 * c**2] = 1
            mo_occ[mo_energy >= lumo] = 0
        if self.verbose >= logger.INFO:
            if mo_energy[n2c+nocc-1]+1e-3 > mo_energy[n2c+nocc]:
                logger.warn(self, 'HOMO %.15g == LUMO %.15g',
                            mo_energy[n2c+nocc-1], mo_energy[n2c+nocc])
            else:
                logger.info(self, 'HOMO %d = %.12g  LUMO %d = %.12g',
                            nocc, mo_energy[n2c+nocc-1],
                            nocc+1, mo_energy[n2c+nocc])
                logger.debug1(self, 'NES  mo_energy = %s', mo_energy[:n2c])
                logger.debug(self, 'PES  mo_energy = %s', mo_energy[n2c:])
        return mo_occ

    make_rdm1 = lib.module_method(make_rdm1, absences=['mo_coeff', 'mo_occ'])
    energy_elec = energy_elec

    def init_direct_scf(self, mol=None):
        if mol is None: mol = self.mol
        def set_vkscreen(opt, name):
            opt._this.r_vkscreen = _vhf._fpointer(name)

        cpu0 = (logger.process_clock(), logger.perf_counter())
        opt_llll = _VHFOpt(mol, 'int2e_spinor', 'CVHFrkbllll_prescreen',
                           'CVHFrkb_q_cond', 'CVHFrkb_dm_cond',
                           direct_scf_tol=self.direct_scf_tol)
        set_vkscreen(opt_llll, 'CVHFrkbllll_vkscreen')

        c1 = .5 / lib.param.LIGHT_SPEED
        opt_ssss = _VHFOpt(mol, 'int2e_spsp1spsp2_spinor',
                           'CVHFrkbllll_prescreen', 'CVHFrkb_q_cond',
                           'CVHFrkb_dm_cond',
                           direct_scf_tol=self.direct_scf_tol/c1**4)
        opt_ssss.direct_scf_tol = self.direct_scf_tol
        opt_ssss.q_cond *= c1**2
        set_vkscreen(opt_ssss, 'CVHFrkbllll_vkscreen')

        opt_ssll = _VHFOpt(mol, 'int2e_spsp1_spinor',
                           'CVHFrkbssll_prescreen',
                           dmcondname='CVHFrkbssll_dm_cond',
                           direct_scf_tol=self.direct_scf_tol)
        opt_ssll.q_cond = numpy.array([opt_llll.q_cond, opt_ssss.q_cond])
        set_vkscreen(opt_ssll, 'CVHFrkbssll_vkscreen')
        logger.timer(self, 'init_direct_scf_coulomb', *cpu0)

        opt_gaunt_lsls = None
        opt_gaunt_lssl = None

        #TODO: prescreen for gaunt
        if self.with_gaunt:
            if self.with_breit:
                # integral function int2e_breit_ssp1ssp2_spinor evaluates
                # -1/2[alpha1*alpha2/r12 + (alpha1*r12)(alpha2*r12)/r12^3]
                intor_prefix = 'int2e_breit_'
            else:
                # integral function int2e_ssp1ssp2_spinor evaluates only
                # alpha1*alpha2/r12. Minus sign was not included.
                intor_prefix = 'int2e_'
            opt_gaunt_lsls = _VHFOpt(mol, intor_prefix + 'ssp1ssp2_spinor',
                                 'CVHFrkb_gaunt_lsls_prescreen', 'CVHFrkb_asym_q_cond',
                                 'CVHFrkb_dm_cond',
                                 direct_scf_tol=self.direct_scf_tol/c1**2)

            opt_gaunt_lssl = _VHFOpt(mol, intor_prefix + 'ssp1sps2_spinor',
                                 'CVHFrkb_gaunt_lssl_prescreen', 'CVHFrkb_asym_q_cond',
                                 'CVHFrkb_dm_cond',
                                 direct_scf_tol=self.direct_scf_tol/c1**2)

            logger.timer(self, 'init_direct_scf_gaunt_breit', *cpu0)
        #return None, None, None, None, None
        return opt_llll, opt_ssll, opt_ssss, opt_gaunt_lsls, opt_gaunt_lssl

    def get_jk(self, mol=None, dm=None, hermi=1, with_j=True, with_k=True,
               omega=None):
        if mol is None: mol = self.mol
        if dm is None: dm = self.make_rdm1()
        t0 = (logger.process_clock(), logger.perf_counter())
        log = logger.new_logger(self)
        if self.direct_scf and self._opt.get(omega) is None:
            with mol.with_range_coulomb(omega):
                self._opt[omega] = self.init_direct_scf(mol)
        vhfopt = self._opt.get(omega)
        if vhfopt is None:
            opt_llll = opt_ssll = opt_ssss = opt_gaunt_lsls = opt_gaunt_lssl = None
        else:
            opt_llll, opt_ssll, opt_ssss, opt_gaunt_lsls, opt_gaunt_lssl = vhfopt
        if self.screening is False:
            opt_llll = opt_ssll = opt_ssss = opt_gaunt_lsls = opt_gaunt_lssl = None

        opt_gaunt = (opt_gaunt_lsls, opt_gaunt_lssl)
        vj, vk = get_jk_coulomb(mol, dm, hermi, self._coulomb_level,
                                opt_llll, opt_ssll, opt_ssss, omega, log)
        t1 = log.timer_debug1('Coulomb', *t0)
        if self.with_breit:
            assert omega is None
            if ('SSSS' in self._coulomb_level.upper() or
                # for the case both with_breit and with_ssss are set
                (not self.with_ssss and 'SSLL' in self._coulomb_level.upper())):
                vj1, vk1 = _call_veff_gaunt_breit(mol, dm, hermi, opt_gaunt, True)
                log.debug('Add Breit term')
                vj += vj1
                vk += vk1
        elif self.with_gaunt and 'SS' in self._coulomb_level.upper():
            assert omega is None
            log.debug('Add Gaunt term')
            vj1, vk1 = _call_veff_gaunt_breit(mol, dm, hermi, opt_gaunt, False)
            vj += vj1
            vk += vk1
        log.timer_debug1('Gaunt and Breit term', *t1)

        log.timer('vj and vk', *t0)
        return vj, vk

    def get_veff(self, mol=None, dm=None, dm_last=0, vhf_last=0, hermi=1):
        '''Dirac-Coulomb'''
        if mol is None: mol = self.mol
        if dm is None: dm = self.make_rdm1()
        if self.direct_scf:
            ddm = numpy.array(dm) - numpy.array(dm_last)
            vj, vk = self.get_jk(mol, ddm, hermi=hermi)
            return numpy.array(vhf_last) + vj - vk
        else:
            vj, vk = self.get_jk(mol, dm, hermi=hermi)
            return vj - vk

    def scf(self, dm0=None):
        cput0 = (logger.process_clock(), logger.perf_counter())

        self.build()
        self.dump_flags()

        if dm0 is None and self.mo_coeff is not None and self.mo_occ is not None:
            # Initial guess from existing wavefunction
            dm0 = self.make_rdm1()

        self.converged, self.e_tot, \
                self.mo_energy, self.mo_coeff, self.mo_occ \
                = kernel(self, self.conv_tol, self.conv_tol_grad,
                         dm0=dm0, callback=self.callback,
                         conv_check=self.conv_check)

        logger.timer(self, 'SCF', *cput0)
        self._finalize()
        return self.e_tot

    def analyze(self, verbose=None):
        if verbose is None: verbose = self.verbose
        return analyze(self, verbose)

    @lib.with_doc(mulliken_pop.__doc__)
    def mulliken_pop(self, mol=None, dm=None, s=None, verbose=logger.DEBUG):
        if mol is None: mol = self.mol
        if dm is None: dm = self.make_rdm1()
        if s is None: s = self.get_ovlp(mol)
        return mulliken_pop(mol, dm, s=s, verbose=verbose)

    @lib.with_doc(dip_moment.__doc__)
    def dip_moment(self, mol=None, dm=None, unit='Debye', verbose=logger.NOTE,
                   **kwargs):
        if mol is None: mol = self.mol
        if dm is None: dm = self.make_rdm1()
        return dip_moment(mol, dm, unit, verbose=verbose, **kwargs)

    def sfx2c1e(self):
        raise NotImplementedError
    def x2c1e(self):
        from pyscf.x2c import x2c
        x2chf = x2c.UHF(self.mol)
        x2chf.__dict__.update(self.__dict__)
        x2chf.mo_energy = None
        x2chf.mo_coeff = None
        x2chf.mo_occ = None
        return x2chf
    x2c = x2c1e

    def nuc_grad_method(self):
        from pyscf.grad import dhf
        return dhf.Gradients(self)

    def reset(self, mol=None):
        '''Reset mol and clean up relevant attributes for scanner mode'''
        if mol is not None:
            self.mol = mol
        self._coulomb_level = 'SSSS' # 'SSSS' ~ LLLL+LLSS+SSSS
        self._opt = {None: None}
        return self

    def stability(self, internal=None, external=None, verbose=None, return_status=False, **kwargs):
        '''
        DHF/DKS stability analysis.

        See also pyscf.scf.stability.rhf_stability function.

        Kwargs:
            return_status: bool
                Whether to return `stable_i` and `stable_e`

        Returns:
            If return_status is False (default), the return value includes
            two set of orbitals, which are more close to the stable condition.
            The first corresponds to the internal stability
            and the second corresponds to the external stability.

            Else, another two boolean variables (indicating current status:
            stable or unstable) are returned.
            The first corresponds to the internal stability
            and the second corresponds to the external stability.
        '''
        from pyscf.scf.stability import dhf_stability
        return dhf_stability(self, verbose, return_status, **kwargs)

    def to_rhf(self):
        raise RuntimeError

    def to_uhf(self):
        raise RuntimeError

    def to_ghf(self):
        raise RuntimeError

    def to_rks(self, xc=None):
        raise RuntimeError

    def to_uks(self, xc=None):
        raise RuntimeError

    def to_gks(self, xc=None):
        raise RuntimeError

    def to_dhf(self):
        return self

    def to_dks(self, xc='HF'):
        '''Convert the input mean-field object to a DKS object.

        Note this conversion only changes the class of the mean-field object.
        The total energy and wave-function are the same as them in the input
        mean-field object.
        '''
        from pyscf.dft import dks
        mf = self.view(dks.UDKS)
        mf.xc = xc
        mf.converged = False
        return mf

    to_ks = to_dks

    to_gpu = lib.to_gpu

UHF = UDHF = DHF


class HF1e(DHF):
    scf = hf._hf1e_scf

    def _eigh(self, h, s):
        if zquatev:
            return zquatev.solve_KR_FCSCE(self.mol, h, s)
        else:
            return DHF._eigh(self, h, s)


class RDHF(DHF):
    '''Kramers restricted Dirac-Hartree-Fock'''
    def __init__(self, mol):
        if mol.nelectron.__mod__(2) != 0:
            raise ValueError('Invalid electron number %i.' % mol.nelectron)
        if zquatev is None:
            raise RuntimeError('zquatev library is required to perform Kramers-restricted DHF')
        UHF.__init__(self, mol)

    def _eigh(self, h, s):
        return zquatev.solve_KR_FCSCE(self.mol, h, s)

    def x2c1e(self):
        from pyscf.x2c import x2c
        x2chf = x2c.RHF(self.mol)
        x2chf.__dict__.update(self.__dict__)
        return x2chf
    x2c = x2c1e

    def to_dks(self, xc='HF'):
        '''Convert the input mean-field object to a DKS object.

        Note this conversion only changes the class of the mean-field object.
        The total energy and wave-function are the same as them in the input
        mean-field object.
        '''
        from pyscf.dft import dks
        mf = self.view(dks.RDKS)
        mf.xc = xc
        mf.converged = False
        return mf

RHF = RDHF


def _ensure_time_reversal_symmetry(mol, mat):
    if mat.ndim == 2:
        mat = [mat]
    for m in mat:
        if abs(m - time_reversal_matrix(mol, m)).max() > 1e-9:
            raise RuntimeError('Matrix does have time reversal symmetry')

def _time_reversal_triu_(mol, vj):
    n2c = vj.shape[1]
    idx, idy = numpy.triu_indices(n2c, 1)
    if vj.ndim == 2:
        Tvj = time_reversal_matrix(mol, vj)
        vj[idx,idy] = Tvj[idy,idx].conj()
    else:
        for i in range(vj.shape[0]):
            Tvj = time_reversal_matrix(mol, vj[i])
            vj[i,idx,idy] = Tvj[idy,idx].conj()
    return vj

def _mat_hermi_(vk, hermi):
    if hermi == 1:
        if vk.ndim == 2:
            vk = lib.hermi_triu(vk, hermi)
        else:
            for i in range(vk.shape[0]):
                vk[i] = lib.hermi_triu(vk[i], hermi)
    return vk

def _jk_triu_(mol, vj, vk, hermi):
    if hermi == 0:
        return _time_reversal_triu_(mol, vj), vk
    else:
        return _mat_hermi_(vj, hermi), _mat_hermi_(vk, hermi)


def _call_veff_llll(mol, dm, hermi=1, mf_opt=None):
    if isinstance(dm, numpy.ndarray) and dm.ndim == 2:
        n2c = dm.shape[0] // 2
        dms = dm[:n2c,:n2c].copy()
    else:
        n2c = dm.shape[1] // 2
        dms = dm[:,:n2c,:n2c].copy()
    vj, vk = _vhf.rdirect_mapdm('int2e_spinor', 's8',
                                ('ji->s2kl', 'jk->s1il'), dms, 1,
                                mol._atm, mol._bas, mol._env, mf_opt)
    return _jk_triu_(mol, vj, vk, hermi)

def _call_veff_ssll(mol, dm, hermi=1, mf_opt=None):
    if isinstance(dm, numpy.ndarray) and dm.ndim == 2:
        dm1 = dm[numpy.newaxis]
    else:
        dm1 = numpy.asarray(dm)
    n_dm = len(dm1)
    n2c = dm1.shape[1] // 2
    dms = numpy.vstack([dm1[:,:n2c,:n2c],
                        dm1[:,n2c:,n2c:],
                        dm1[:,n2c:,:n2c],
                        dm1[:,:n2c,n2c:]])
    if hermi:
        jks = (['lk->s2ij'] * n_dm +
               ['ji->s2kl'] * n_dm +
               ['jk->s1il'] * n_dm)
    else:
        jks = (['lk->s2ij'] * n_dm +
               ['ji->s2kl'] * n_dm +
               ['jk->s1il'] * n_dm +
               ['li->s1kj'] * n_dm)
    c1 = .5 / lib.param.LIGHT_SPEED
    vx = _vhf.rdirect_bindm('int2e_spsp1_spinor', 's4', jks, dms, 1,
                            mol._atm, mol._bas, mol._env, mf_opt)
    vx = vx.reshape(-1,n_dm,n2c,n2c) * c1**2
    vj = numpy.zeros((n_dm,n2c*2,n2c*2), dtype=numpy.complex128)
    vk = numpy.zeros((n_dm,n2c*2,n2c*2), dtype=numpy.complex128)
    if hermi == 0:
        vj[:,n2c:,n2c:] = _time_reversal_triu_(mol, vx[0])
        vj[:,:n2c,:n2c] = _time_reversal_triu_(mol, vx[1])
        vk[:,n2c:,:n2c] = vx[2]
        vk[:,:n2c,n2c:] = vx[3]
    else:
        vj[:,n2c:,n2c:] = _mat_hermi_(vx[0], hermi)
        vj[:,:n2c,:n2c] = _mat_hermi_(vx[1], hermi)
        vk[:,n2c:,:n2c] = vx[2]
        vk[:,:n2c,n2c:] = vx[2].conj().transpose(0,2,1)
    vj = vj.reshape(dm.shape)
    vk = vk.reshape(dm.shape)
    return vj, vk

def _call_veff_ssss(mol, dm, hermi=1, mf_opt=None):
    c1 = .5 / lib.param.LIGHT_SPEED
    if isinstance(dm, numpy.ndarray) and dm.ndim == 2:
        n2c = dm.shape[0] // 2
        dms = dm[n2c:,n2c:].copy()
    else:
        n2c = dm[0].shape[0] // 2
        dms = []
        for dmi in dm:
            dms.append(dmi[n2c:,n2c:].copy())
    vj, vk = _vhf.rdirect_mapdm('int2e_spsp1spsp2_spinor', 's8',
                                ('ji->s2kl', 'jk->s1il'), dms, 1,
                                mol._atm, mol._bas, mol._env, mf_opt) * c1**4
    return _jk_triu_(mol, vj, vk, hermi)

def _call_veff_gaunt_breit(mol, dm, hermi=1, mf_opt=None, with_breit=False):
    if mf_opt is not None:
        opt_gaunt_lsls, opt_gaunt_lssl = mf_opt
    else:
        opt_gaunt_lsls = opt_gaunt_lssl = None

    log = logger.new_logger(mol)
    t0 = (logger.process_clock(), logger.perf_counter())

    if with_breit:
        # integral function int2e_breit_ssp1ssp2_spinor evaluates
        # -1/2[alpha1*alpha2/r12 + (alpha1*r12)(alpha2*r12)/r12^3]
        intor_prefix = 'int2e_breit_'
    else:
        # integral function int2e_ssp1ssp2_spinor evaluates only
        # alpha1*alpha2/r12. Minus sign was not included.
        intor_prefix = 'int2e_'

    if hermi == 0 and DEBUG:
        _ensure_time_reversal_symmetry(mol, dm)

    if isinstance(dm, numpy.ndarray) and dm.ndim == 2:
        n_dm = 1
        n2c = dm.shape[0] // 2
        dmls = dm[:n2c,n2c:].copy()
        dmsl = dm[n2c:,:n2c].copy()
        dmll = dm[:n2c,:n2c].copy()
        dmss = dm[n2c:,n2c:].copy()
        dms = [dmsl, dmls, dmll, dmss]
    else:
        n_dm = len(dm)
        n2c = dm[0].shape[0] // 2
        dmll = [dmi[:n2c,:n2c].copy() for dmi in dm]
        dmls = [dmi[:n2c,n2c:].copy() for dmi in dm]
        dmsl = [dmi[n2c:,:n2c].copy() for dmi in dm]
        dmss = [dmi[n2c:,n2c:].copy() for dmi in dm]
        dms = dmsl + dmls + dmll + dmss
    vj = numpy.zeros((n_dm,n2c*2,n2c*2), dtype=numpy.complex128)
    vk = numpy.zeros((n_dm,n2c*2,n2c*2), dtype=numpy.complex128)

    if hermi != 0:
        jks = ('lk->s1ij', 'jk->s1il')
        vj_ls, vk_ls = _vhf.rdirect_mapdm(
            intor_prefix+'ssp1ssp2_spinor', 's1', jks, dms[:n_dm], 1,
            mol._atm, mol._bas, mol._env, opt_gaunt_lsls)
        vj[:,:n2c,n2c:] = vj_ls
        vk[:,:n2c,n2c:] = vk_ls
        t0 = log.timer_debug1('LSLS contribution', *t0)

        jks = (('lk->s1ij',) * n_dm +
               ('li->s1kj',) * n_dm +
               ('jk->s1il',) * n_dm)
        vx = _vhf.rdirect_bindm(
            intor_prefix+'ssp1sps2_spinor', 's1', jks, dms[n_dm:], 1,
            mol._atm, mol._bas, mol._env, opt_gaunt_lssl)
        vj[:,:n2c,n2c:]+= vx[      :n_dm  ,:,:]
        vk[:,n2c:,n2c:] = vx[n_dm  :n_dm*2,:,:]
        vk[:,:n2c,:n2c] = vx[n_dm*2:      ,:,:]
        t0 = log.timer_debug1('LSSL contribution', *t0)

        if hermi == 1:
            vj[:,n2c:,:n2c] = vj[:,:n2c,n2c:].transpose(0,2,1).conj()
            vk[:,n2c:,:n2c] = vk[:,:n2c,n2c:].transpose(0,2,1).conj()
        elif hermi == 2:
            vj[:,n2c:,:n2c] = -vj[:,:n2c,n2c:].transpose(0,2,1).conj()
            vk[:,n2c:,:n2c] = -vk[:,:n2c,n2c:].transpose(0,2,1).conj()
    else:
        jks = ('lk->s1ij', 'jk->s1il')
        subj, subk = _vhf.rdirect_mapdm(
            intor_prefix+'ssp1ssp2_spinor', 's1', jks,
            (dms[:n_dm] + [x.conj().T for x in dms[n_dm:n_dm*2]]), 1,
            mol._atm, mol._bas, mol._env, opt_gaunt_lsls)
        vj[:,:n2c,n2c:] = subj[      :n_dm  ,:,:]
        vj[:,n2c:,:n2c] = subj[n_dm  :n_dm*2,:,:].conj().transpose(0,2,1)
        vk[:,:n2c,n2c:] = subk[      :n_dm  ,:,:]
        vk[:,n2c:,:n2c] = subk[n_dm  :n_dm*2,:,:].conj().transpose(0,2,1)
        t0 = log.timer_debug1('LSLS contribution', *t0)

        jks = (('ji->s1kl',) * n_dm +
               ('lk->s1ij',) * n_dm +
               ('li->s1kj',) * n_dm +
               ('jk->s1il',) * n_dm)
        vx = _vhf.rdirect_bindm(
            intor_prefix+'ssp1sps2_spinor', 's1', jks, dms, 1,
            mol._atm, mol._bas, mol._env, opt_gaunt_lssl)
        vj[:,n2c:,:n2c]+= vx[      :n_dm  ,:,:]
        vj[:,:n2c,n2c:]+= vx[n_dm  :n_dm*2,:,:]
        vk[:,n2c:,n2c:] = vx[n_dm*2:n_dm*3,:,:]
        vk[:,:n2c,:n2c] = vx[n_dm*3:n_dm*4,:,:]
        t0 = log.timer_debug1('LSSL contribution', *t0)

    vj = vj.reshape(dm.shape)
    vk = vk.reshape(dm.shape)
    c1 = .5 / lib.param.LIGHT_SPEED
    if with_breit:
        vj *= c1**2
        vk *= c1**2
    else:
        vj *= -c1**2
        vk *= -c1**2
    return vj, vk

def _proj_dmll(mol_nr, dm_nr, mol):
    '''Project non-relativistic atomic density matrix to large component spinor
    representation
    '''
    from pyscf.scf import addons
    proj = addons.project_mo_nr2r(mol_nr, numpy.eye(mol_nr.nao_nr()), mol)

    n2c = proj.shape[0]
    n4c = n2c * 2
    dm = numpy.zeros((n4c,n4c), dtype=numpy.complex128)
    # *.5 because alpha and beta are summed in project_mo_nr2r
    dm_ll = reduce(numpy.dot, (proj, dm_nr*.5, proj.T.conj()))
    dm[:n2c,:n2c] = (dm_ll + time_reversal_matrix(mol, dm_ll)) * .5
    return dm

class _VHFOpt(_vhf._VHFOpt):
    def set_dm(self, dm, atm, bas, env):
        if self._dmcondname is None:
            return

        mol = self.mol
        if isinstance(dm, numpy.ndarray) and dm.ndim == 2:
            n_dm = 1
        else:
            n_dm = len(dm)
        dm = numpy.asarray(dm, order='C')
        ao_loc = mol.ao_loc_2c()
        if isinstance(self._dmcondname, ctypes._CFuncPtr):
            fdmcond = self._dmcondname
        else:
            fdmcond = getattr(_vhf.libcvhf, self._dmcondname)
        nbas = mol.nbas
        dm_cond = numpy.empty((n_dm*2, nbas, nbas))
        fdmcond(dm_cond.ctypes, dm.ctypes, ctypes.c_int(n_dm),
                ao_loc.ctypes, mol._atm.ctypes, ctypes.c_int(mol.natm),
                mol._bas.ctypes, ctypes.c_int(nbas), mol._env.ctypes)
        self.dm_cond = dm_cond


if __name__ == '__main__':
    import pyscf.gto
    mol = pyscf.gto.Mole()
    mol.verbose = 5
    mol.output = 'out_dhf'

    mol.atom.extend([['He', (0.,0.,0.)], ])
    mol.basis = {
        'He': [(0, 0, (1, 1)),
               (0, 0, (3, 1)),
               (1, 0, (1, 1)), ]}
    mol.build()

    ##############
    # SCF result
    method = UHF(mol)
    energy = method.scf() #-2.38146942868
    print(energy)
    method.with_gaunt = True
    print(method.scf()) # -2.38138339005
    method.with_breit = True
    print(method.scf()) # -2.38138339005
